import logging
import re

from multiqc import config
from multiqc.modules.base_module import BaseMultiqcModule, ModuleNoSamplesFound
from multiqc.plots import bargraph

log = logging.getLogger(__name__)


class MultiqcModule(BaseMultiqcModule):
    def __init__(self):
        # Initialise the parent object
        super(MultiqcModule, self).__init__(
            name="Porechop",
            anchor="porechop",
            href="https://github.com/rrwick/Porechop",
            info="a tool for finding and removing adapters from Oxford Nanopore reads.",
            # doi="",  # No DOI available
        )

        # Find and load reports
        self.porechop_data = dict()

        # Find all files for porechop
        for f in self.find_log_files("porechop", filehandles=True):
            self.parse_logs(f)

        self.porechop_data = self.ignore_samples(self.porechop_data)

        if len(self.porechop_data) == 0:
            raise ModuleNoSamplesFound

        log.info(f"Found {len(self.porechop_data)} reports")

        # Superfluous function call to confirm that it is used in this module
        # Replace None with actual version if it is available
        self.add_software_version(None)

        # Write data to file
        self.write_data_file(self.porechop_data, "porechop")

        self.porechop_general_stats()
        if max(len(v) for v in self.porechop_data.values()) > 1:
            self.start_trim_barplot()
            self.end_trim_barplot()
            self.middle_split_barplot()
        self.no_adapters_found()

    def parse_logs(self, f):
        """Parsing Logs. Note: careful of ANSI formatting log"""

        def get_float(val):
            """Get float from string"""
            val = val.replace(",", "")
            try:
                return float(val)
            except ValueError:
                return val

        for line in f["f"]:
            # Find line after loading reads, and remove suffixes for sample name
            if "Loading reads" in line:
                s_name = next(f["f"]).rstrip()
                s_name = self.clean_s_name(s_name, f)
                self.add_data_source(f, s_name=s_name)
                if s_name in self.porechop_data:
                    log.debug(f"Duplicate sample name found! Overwriting: {s_name}")
                self.porechop_data[s_name] = {}

            # Find each valid metric, clean up for plain integer
            # 10,000 reads loaded
            if "reads loaded" in line:
                reads_loaded = re.search(r"([\d,]+)\s*reads loaded", line)
                if reads_loaded:
                    self.porechop_data[s_name]["Input Reads"] = get_float(reads_loaded.group(1))

            # 7,100 / 10,000 reads had adapters trimmed from their start (425,196 bp removed)
            if "reads had adapters trimmed from their start" in line:
                adapter_start = re.search(
                    r"([\d,]+)\s*/\s*([\d,]+)\s*reads had adapters trimmed from their start \(([\d,]+) bp removed\)",
                    line,
                )
                if adapter_start:
                    self.porechop_data[s_name]["Start Trimmed"] = get_float(adapter_start.group(1))
                    self.porechop_data[s_name]["Start Trimmed Total"] = get_float(adapter_start.group(2))
                    self.porechop_data[s_name]["Start Trimmed (bp)"] = get_float(adapter_start.group(3))
                    self.porechop_data[s_name]["Start Untrimmed"] = (
                        self.porechop_data[s_name]["Start Trimmed Total"] - self.porechop_data[s_name]["Start Trimmed"]
                    )
                    try:
                        self.porechop_data[s_name]["Start Trimmed Percent"] = (
                            self.porechop_data[s_name]["Start Trimmed"]
                            / self.porechop_data[s_name]["Start Trimmed Total"]
                        ) * 100
                    except ZeroDivisionError:
                        pass

            # 4,849 / 10,000 reads had adapters trimmed from their end (283,192 bp removed)'
            if "reads had adapters trimmed from their end" in line:
                end_trimmed = re.search(
                    r"([\d,]+)\s*/\s*([\d,]+)\s*reads had adapters trimmed from their end \(([\d,]+) bp removed\)", line
                )
                if end_trimmed:
                    self.porechop_data[s_name]["End Trimmed"] = get_float(end_trimmed.group(1))
                    self.porechop_data[s_name]["End Trimmed Total"] = get_float(end_trimmed.group(2))
                    self.porechop_data[s_name]["End Trimmed (bp)"] = get_float(end_trimmed.group(3))
                    self.porechop_data[s_name]["End Untrimmed"] = (
                        self.porechop_data[s_name]["End Trimmed Total"] - self.porechop_data[s_name]["End Trimmed"]
                    )
                    try:
                        self.porechop_data[s_name]["End Trimmed Percent"] = (
                            self.porechop_data[s_name]["End Trimmed"]
                            / self.porechop_data[s_name]["Start Trimmed Total"]
                            * 100
                        )
                    except ZeroDivisionError:
                        pass

            # 7 / 10,000 reads were split based on middle adapters
            if "reads were split based on middle adapters" in line:
                split_stats = re.search(r"([\d,]+)\s*/\s*([\d,]+)\s*reads were split based on middle adapters", line)
                if split_stats:
                    self.porechop_data[s_name]["Middle Split"] = get_float(split_stats.group(1))
                    self.porechop_data[s_name]["Middle Split Total"] = get_float(split_stats.group(2))
                    self.porechop_data[s_name]["Middle Not-Split"] = (
                        self.porechop_data[s_name]["Middle Split Total"] - self.porechop_data[s_name]["Middle Split"]
                    )
                    try:
                        self.porechop_data[s_name]["Middle Split Percent"] = (
                            self.porechop_data[s_name]["Middle Split"]
                            / self.porechop_data[s_name]["Middle Split Total"]
                            * 100
                        )
                    except ZeroDivisionError:
                        pass

    def porechop_general_stats(self):
        """Porechop General Stats Table"""
        headers = {
            "Input Reads": {
                "title": f"Input Reads ({config.read_count_prefix})",
                "description": f"Number of reads loaded into Porechop ({config.read_count_prefix})",
                "scale": "Greens",
                "shared_key": "read_count",
                "modify": lambda x: x * config.read_count_multiplier,
            },
            "Start Trimmed": {
                "title": f"Start Trimmed ({config.read_count_prefix})",
                "description": "Number of reads that had adapters trimmed from the start ({})".format(
                    config.read_count_prefix
                ),
                "scale": "Purples",
                "shared_key": "read_count",
                "modify": lambda x: x * config.read_count_multiplier,
                "hidden": True,
            },
            "Start Trimmed Percent": {
                "title": "Start Trimmed",
                "description": "Percent of reads that had adapters trimmed from the start",
                "suffix": "%",
                "max": 100,
                "scale": "RdYlGn",
            },
            "End Trimmed": {
                "title": f"End Trimmed ({config.read_count_prefix})",
                "description": "Number of reads that had adapters trimmed from the end ({})".format(
                    config.read_count_prefix
                ),
                "scale": "Purples",
                "shared_key": "read_count",
                "modify": lambda x: x * config.read_count_multiplier,
                "hidden": True,
            },
            "End Trimmed Percent": {
                "title": "End Trimmed",
                "description": "Percent of reads that had adapters trimmed from the end",
                "suffix": "%",
                "max": 100,
                "scale": "RdYlGn",
            },
            "Middle Split": {
                "title": f"Middle Split ({config.read_count_prefix})",
                "description": f"Number of reads split based on middle adapters ({config.read_count_prefix})",
                "scale": "Purples",
                "shared_key": "read_count",
                "modify": lambda x: x * config.read_count_multiplier,
                "hidden": True,
            },
            "Middle Split Percent": {
                "title": "Middle Split",
                "description": "Percent of reads that were split based on middle adapters",
                "suffix": "%",
                "max": 100,
                "scale": "RdYlGn",
            },
        }

        self.general_stats_addcols(self.porechop_data, headers)

    def start_trim_barplot(self):
        """Barplot of number of reads adapter trimmed at read start"""
        cats = {
            "Start Trimmed": {"name": "Start Trimmed", "color": "#7cb5ec"},
            "Start Untrimmed": {"name": "Start Untrimmed", "color": "#f7a35c"},
        }
        config = {
            "id": "porechop-starttrim-barplot",
            "title": "Porechop: Read Start Adapter Timmed",
            "ylab": "Read Counts",
        }
        self.add_section(
            name="Reads adapter-trimmed read start",
            anchor="porechop-starttrim",
            description="Shows the number of reads that had adapters removed from read start.",
            plot=bargraph.plot(self.porechop_data, cats, config),
        )

    def end_trim_barplot(self):
        """Barplot of number of reads adapter trimmed at read end"""
        cats = {
            "End Trimmed": {"name": "End Trimmed", "color": "#7cb5ec"},
            "End Untrimmed": {"name": "End Untrimmed", "color": "#f7a35c"},
        }
        config = {
            "id": "porechop-endtrim-barplot",
            "title": "Porechop: Read End Adapter Timmed",
            "ylab": "Read Counts",
        }
        self.add_section(
            name="Reads adapter-trimmed read end",
            anchor="porechop-endtrim",
            description="Shows the number of reads that had adapters removed from read end.",
            plot=bargraph.plot(self.porechop_data, cats, config),
        )

    def middle_split_barplot(self):
        """Barplot of number of reads adapter trimmed at read end"""
        cats = {
            "Middle Split": {"name": "Split Reads", "color": "#7cb5ec"},
            "Middle Not-Split": {"name": "Unsplit Reads", "color": "#f7a35c"},
        }
        config = {
            "id": "porechop-middlesplit-barplot",
            "title": "Porechop: Middle Split",
            "ylab": "Read Counts",
        }
        self.add_section(
            name="Middle split reads",
            anchor="porechop-middlesplit",
            description="Shows the number of reads that were split due to adapter being present in middle of read.",
            plot=bargraph.plot(self.porechop_data, cats, config),
        )

    def no_adapters_found(self):
        """Show any samples that did not have any trimming"""
        no_adapters = []
        for s_name in self.porechop_data:
            if len(self.porechop_data[s_name]) == 1:
                no_adapters.append(s_name)
        if len(no_adapters):
            self.add_section(
                name="No adapters found",
                anchor="porechop-noadapters",
                description="The following samples did not have any adapters found - output reads were unchanged from input reads:",
                content=f"""
                    <ul>
                        <li><code>{'</code></li><li><code>'.join(no_adapters)}</code></li>
                    </ul>
                """,
            )



import logging
from collections import defaultdict

from multiqc.modules.base_module import BaseMultiqcModule
from multiqc.plots import linegraph

from .util import average_from_range, percentage_from_content_metric

log = logging.getLogger(__name__)

N_QV = 2
ADAPTER_SEQS = ["AGATCGGAAGAG", "ATGGAATTCTCG", "CTGTCTCTTATA"]


class DragenFastqcGcMetrics(BaseMultiqcModule):
    """
    Rendering ALL THE THINGS!

    Not to be confused with DragenGcMetrics
    """

    def add_gc_metrics(self):
        # Now add each section in order
        self.gc_content_plot()
        self.gc_content_mean_quality_plot()
        self.get_avg_gc_content_by_sample()

        return self.dragen_fastqc_data.keys()

    def gc_content_plot(self):
        """Create the HTML for the FastQC GC content plot"""

        data = dict()
        data_norm = dict()
        LEN_GROUP = "READ LENGTHS"
        GC_GROUP = "READ GC CONTENT"
        for s_name in sorted(self.dragen_fastqc_data):
            for mate in sorted(self.dragen_fastqc_data[s_name]):
                r_name = f"{s_name}_{mate}"

                # First figure out the baseline
                max_len = 0
                for metric, value in self.dragen_fastqc_data[s_name][mate][LEN_GROUP].items():
                    if int(value) > 0:
                        pos = average_from_range(metric.split()[0][:-2])
                        max_len = max(pos, max_len)

                data[r_name] = defaultdict(float)
                for metric, value in self.dragen_fastqc_data[s_name][mate][GC_GROUP].items():
                    pct = percentage_from_content_metric(metric)
                    data[r_name][pct] = value

                data_norm[r_name] = dict()
                total = sum([c for c in data[r_name].values()])
                for gc, count in data[r_name].items():
                    if total > 0:
                        data_norm[r_name][gc] = (count / total) * 100

        if len(data) == 0:
            log.debug("per_sequence_gc_content not found in FastQC reports")
            return None

        pconfig = {
            "id": "dragenqc_per_sequence_gc_content_plot",
            "title": "DRAGEN-QC: Per-Sequence GC Content",
            "xlab": "% GC",
            "ylab": "Percentage",
            "ymin": 0,
            "xmax": 100,
            "xmin": 0,
            "yDecimals": False,
            "tt_label": "<b>{point.x}% GC</b>: {point.y:.2f}",
            # 'colors': self.get_status_cols('per_sequence_gc_content'),
            "data_labels": [{"name": "Percentages", "ylab": "Percentage"}, {"name": "Counts", "ylab": "Count"}],
        }

        self.add_section(
            name="Per-Sequence GC Content",
            anchor="dragenqc_per_sequence_gc_content",
            description="""The average GC content of reads. Normal random library typically have a
            roughly normal distribution of GC content.""",
            helptext="""
            From the [FastQC help](http://www.bioinformatics.babraham.ac.uk/projects/fastqc/Help/3%20Analysis%20Modules/5%20Per%20Sequence%20GC%20Content.html):
            _This module measures the GC content across the whole length of each sequence
            in a file and compares it to a modelled normal distribution of GC content._
            _In a normal random library you would expect to see a roughly normal distribution
            of GC content where the central peak corresponds to the overall GC content of
            the underlying genome. Since we don't know the the GC content of the genome the
            modal GC content is calculated from the observed data and used to build a
            reference distribution._
            _An unusually shaped distribution could indicate a contaminated library or
            some other kinds of biased subset. A normal distribution which is shifted
            indicates some systematic bias which is independent of base position. If there
            is a systematic bias which creates a shifted normal distribution then this won't
            be flagged as an error by the module since it doesn't know what your genome's
            GC content should be._
            """,
            plot=linegraph.plot([data_norm, data], pconfig),
        )

    def gc_content_mean_quality_plot(self):
        """Create the HTML for the positional mean-quality score plot"""

        GROUP = "READ GC CONTENT QUALITY"
        data = dict()
        for s_name in sorted(self.dragen_fastqc_data):
            for mate in sorted(self.dragen_fastqc_data[s_name]):
                r_name = f"{s_name}_{mate}"
                data[r_name] = dict()

                for key, value in self.dragen_fastqc_data[s_name][mate][GROUP].items():
                    parts = key.split()
                    pct = int(parts[0][:-1])
                    try:
                        data[r_name][pct] = float(value)
                    except Exception:
                        continue

        pconfig = {
            "id": "fastqc_gc_content_mean_sequence_quality_plot",
            "title": "DRAGEN-QC: GC Content Mean Quality Scores",
            "ylab": "Phred Quality Score",
            "xlab": "% GC",
            "ymin": 0,
            "xDecimals": False,
            "tt_label": "<b>Base {point.x}</b>: {point.y:.2f}",
            # 'colors': self.get_status_cols('per_base_sequence_quality'),
            "yPlotBands": [
                {"from": 28, "to": 100, "color": "#c3e6c3"},
                {"from": 20, "to": 28, "color": "#e6dcc3"},
                {"from": 0, "to": 20, "color": "#e6c3c3"},
            ],
        }

        self.add_section(
            name="GC Content Mean Quality Scores",
            anchor="fastqc_gc_content_mean_sequence_quality",
            description="The mean quality value across each base position in the read.",
            helptext="""
            To enable multiple samples to be plotted on the same graph, only the mean quality
            scores are plotted (unlike the box plots seen in FastQC reports).

            Taken from the [FastQC help](http://www.bioinformatics.babraham.ac.uk/projects/fastqc/Help/3%20Analysis%20Modules/2%20Per%20Base%20Sequence%20Quality.html):

            _The y-axis on the graph shows the quality scores. The higher the score, the better
            the base call. The background of the graph divides the y axis into very good quality
            calls (green), calls of reasonable quality (orange), and calls of poor quality (red).
            The quality of calls on most platforms will degrade as the run progresses, so it is
            common to see base calls falling into the orange area towards the end of a read._
            """,
            plot=linegraph.plot(data, pconfig),
        )

    def get_avg_gc_content_by_sample(self):
        data = dict()
        avg_gc_content_data = dict()
        GC_GROUP = "READ GC CONTENT"
        for s_name in sorted(self.dragen_fastqc_data):
            for mate in sorted(self.dragen_fastqc_data[s_name]):
                r_name = f"{s_name}_{mate}"
                data[r_name] = defaultdict(float)
                for metric, value in self.dragen_fastqc_data[s_name][mate][GC_GROUP].items():
                    pct = percentage_from_content_metric(metric)
                    data[r_name][pct] = value

            reads_by_gc_sum = 0
            total_reads_sum = 0
            for mate, gc_data in data.items():
                for pct, num_reads in gc_data.items():
                    reads_by_gc_sum += (1.0 * pct / 100) * num_reads
                    total_reads_sum += num_reads

            avg_gc_content_data[s_name] = {"avg_gc_content_percent": (reads_by_gc_sum * 100) / total_reads_sum}

        # Add Avg. GC Content to header
        headers = {
            "avg_gc_content_percent": {
                "title": "% GC",
                "description": "Average % GC Content",
                "max": 100,
                "min": 0,
                "suffix": "%",
                "scale": "PuRd",
                "format": "{:,.0f}",
            }
        }
        self.general_stats_addcols(avg_gc_content_data, headers, namespace="FastQC")



""" MultiQC module to parse output from MEGAHIT """


import logging
import re


from multiqc import config
from multiqc.modules.base_module import BaseMultiqcModule, ModuleNoSamplesFound
from multiqc.plots import table

# Initialise the logger
log = logging.getLogger(__name__)


class MultiqcModule(BaseMultiqcModule):
    """MEGAHIT module"""

    def __init__(self):
        # Initialise the parent object
        super(MultiqcModule, self).__init__(
            name="MEGAHIT",
            anchor="megahit",
            href="https://github.com/voutcn/megahit",
            info="is an ultra-fast and memory-efficient NGS assembler",
            doi="10.1093/bioinformatics/btv033",
        )

        data = dict()
        for f in self.find_log_files("megahit", filehandles=True):
            for line in f["f"]:
                self.add_data_source(f, f["s_name"])
                if " - MEGAHIT v" in line:
                    version = line.split(" - MEGAHIT v")[1].strip()
                    self.add_software_version(version, sample=f["s_name"])
                elif " - Memory used: " in line:
                    mem = line.split(" - Memory used: ")[1].strip()
                    try:
                        mem_mb = float(mem) / 1024 / 1024
                    except ValueError:
                        mem_mb = None
                    data[f["s_name"]] = {"megahit_mem": mem_mb}
                elif " - ALL DONE. Time elapsed: " in line:
                    time = line.split(" - ALL DONE. Time elapsed: ")[1].strip().replace(" seconds", "")
                    data[f["s_name"]]["megahit_time"] = time
                elif " contigs, total " in line:
                    # Parse line like this:
                    # 2022-06-13 14:22:47 - 64408 contigs, total 46322050 bp, min 200 bp, max 94571 bp, avg 719 bp, N50 831 bp
                    regex = re.compile(
                        r"(\d+) contigs, total (\d+) bp, min (\d+) bp, max (\d+) bp, avg (\d+) bp, N50 (\d+) bp"
                    )
                    match = regex.search(line)
                    if match:
                        data[f["s_name"]]["megahit_contigs"] = int(match.group(1))
                        data[f["s_name"]]["megahit_bases"] = int(match.group(2))
                        data[f["s_name"]]["megahit_min_contig"] = int(match.group(3))
                        data[f["s_name"]]["megahit_max_contig"] = int(match.group(4))
                        data[f["s_name"]]["megahit_avg_contig"] = int(match.group(5))
                        data[f["s_name"]]["megahit_n50"] = int(match.group(6))

        # Filter to strip out ignored sample names
        data = self.ignore_samples(data)
        if len(data) == 0:
            raise ModuleNoSamplesFound
        log.info(f"Found {len(data)} reports")

        # Write parsed report data to a file
        self.write_data_file(data, f"multiqc_{self.anchor}")

        headers = {
            "megahit_contigs": {
                "title": "# Contigs",
                "description": "Number of contigs",
                "min": 0,
                "scale": "Purples",
                "format": "{:,d}",
            },
            "megahit_bases": {
                "title": f"Bases, {config.base_count_prefix}",
                "description": f"Total number of bases ({config.base_count_desc})",
                "shared_key": "base_count",
                "min": 0,
                "scale": "Blues",
            },
            "megahit_min_contig": {
                "title": "Min contig length",
                "description": "Minimum contig length",
                "min": 0,
                "scale": "YlGn",
                "suffix": " bp",
                "format": "{:,d}",
            },
            "megahit_max_contig": {
                "title": "Max contig length",
                "description": "Maximum contig length",
                "min": 0,
                "scale": "YlGn",
                "suffix": " bp",
                "format": "{:,d}",
            },
            "megahit_avg_contig": {
                "title": "Avg contig length",
                "description": "Average contig length",
                "min": 0,
                "scale": "YlGn",
                "suffix": " bp",
                "format": "{:,d}",
            },
            "megahit_n50": {
                "title": "N50",
                "description": "N50 contig length",
                "min": 0,
                "scale": "YlGnBu",
                "suffix": " bp",
                "format": "{:,d}",
            },
            "megahit_mem": {
                "title": "Memory",
                "description": "Memory used, MB",
                "min": 0,
                "suffix": " MB",
                "scale": "Reds",
            },
            "megahit_time": {
                "title": "Time",
                "description": "Time elapsed",
                "min": 0,
                "suffix": " s",
                "scale": "Greys",
            },
        }

        # Basic Stats Table
        genstats_cols = ["megahit_contigs", "megahit_avg_contig", "megahit_n50"]
        self.general_stats_addcols(data, {h: v for h, v in headers.items() if h in genstats_cols})

        # Separate table with all metrics
        self.add_section(
            name="Run statistics",
            anchor="megahit-stats",
            description="Metrics and run statistics from MEGAHIT run logs.",
            plot=table.plot(
                data,
                headers,
                pconfig={"id": "megahit-stats-table", "title": "Run statistics"},
            ),
        )



import logging
import re
from collections import defaultdict

from multiqc.modules.base_module import BaseMultiqcModule
from multiqc.modules.qualimap.QM_BamQC import coverage_histogram_helptext, genome_fraction_helptext
from multiqc.plots import linegraph

# Initialise the logger
log = logging.getLogger(__name__)


class DragenCoverageHist(BaseMultiqcModule):
    def add_coverage_hist(self):
        data_by_phenotype_by_sample = defaultdict(dict)
        for f in self.find_log_files("dragen/wgs_fine_hist"):
            data_by_phenotype = parse_wgs_fine_hist(f)
            s_name = f["s_name"]
            if s_name in data_by_phenotype_by_sample:
                log.debug(f"Duplicate sample name found! Overwriting: {s_name}")
            self.add_data_source(f, section="wgs_fine_hist")
            data_by_phenotype_by_sample[s_name].update(data_by_phenotype)

            # Superfluous function call to confirm that it is used in this module
            # Replace None with actual version if it is available
            self.add_software_version(None, s_name)

        # Filter to strip out ignored sample names:
        data_by_phenotype_by_sample = self.ignore_samples(data_by_phenotype_by_sample)

        # Merge tumor and normal data:
        data_by_sample = defaultdict(dict)
        for sn in data_by_phenotype_by_sample:
            for phenotype in data_by_phenotype_by_sample[sn]:
                new_sn = sn
                if phenotype == "normal":
                    new_sn = sn + "_normal"
                data_by_sample[new_sn] = data_by_phenotype_by_sample[sn][phenotype]
        if not data_by_sample:
            return set()

        # Only plot data, don't want to write this to a file
        # (can do so with --export-plots already)
        # self.write_data_file(data_by_sample, "dragen_cov_hist")

        dist_data = {sn: dist for sn, (dist, cum, depth_1pc) in data_by_sample.items()}
        cum_data = {sn: cum for sn, (dist, cum, depth_1pc) in data_by_sample.items()}
        depth_1pc = max(depth_1pc for (dist, cum, depth_1pc) in data_by_sample.values())

        self.add_section(
            name="Coverage distribution",
            anchor="dragen-coverage-distribution",
            description="Number of locations in the reference genome with a given depth of coverage.",
            helptext=coverage_histogram_helptext,
            plot=linegraph.plot(
                dist_data,
                {
                    "id": "dragen_coverage_dist",
                    "title": "Dragen: Coverage distribution",
                    "xlab": "Depth (x)",
                    "ylab": "Number of bases in genome covered by X reads",
                    "ymin": 0,
                    "xmin": 0,
                    "xmax": depth_1pc,  # trim long flat tail
                    "tt_label": "<b>{point.x}X</b>: {point.y} loci",
                    "cpswitch": True,
                },
            ),
        )

        self.add_section(
            name="Cumulative coverage hist",
            anchor="dragen-cum-coverage-histogram",
            description="Number of locations in the reference genome with at least given depth of coverage.",
            helptext=genome_fraction_helptext,
            plot=linegraph.plot(
                cum_data,
                {
                    "id": "dragen_cumulative_coverage_hist",
                    "title": "Dragen: Cumulative coverage hist",
                    "xlab": "Depth (x)",
                    "ylab": "% of bases in genome covered by at least X reads",
                    "ymin": 0,
                    "ymax": 100,
                    "xmin": 0,
                    "xmax": depth_1pc,  # trim long flat tail
                    "tt_label": "<b>{point.x}X</b>: {point.y:.2f}%",
                },
            ),
        )
        return data_by_sample.keys()


def parse_wgs_fine_hist(f):
    """
    T_SRR7890936_50pc.wgs_fine_hist_normal.csv
    T_SRR7890936_50pc.wgs_fine_hist_tumor.csv

    Depth,Overall
    0,104231614
    1,9430586
    2,5546235
    ...
    998,208
    999,177
    1000+,201801

    Contains two columns: Depth and Overall. The value in the Depth column ranges from 0 to 1000+
    and the Overall column indicates the number of loci covered at the corresponding depth.

    Parsing all values except for 1000+ and plotting a distribution histogram and cumulative histogram
    """

    # first pass to calculate total number of bases to calculate percentages
    parsed_data = dict()
    for line in f["f"].splitlines():
        if line.startswith("Depth,Overall"):
            continue
        key, cnt = line.split(",")
        try:
            cnt = int(cnt)
        except ValueError:
            continue
        parsed_data[key] = cnt

    total_cnt = sum(parsed_data.values())

    data = dict()
    cum_data = dict()
    cum_cnt = 0
    depth_1pc = None

    for key, cnt in reversed(list(parsed_data.items())):
        try:
            depth = int(key)
        except ValueError:
            continue
        cum_cnt += cnt
        if total_cnt > 0:
            cum_pct = cum_cnt / total_cnt * 100.0
        else:
            cum_pct = 0
        if cum_pct < 1:  # to trim long flat tail
            depth_1pc = depth
        data[depth] = cnt
        cum_data[depth] = cum_pct

    m = re.search(r"(tumor|normal).csv", f["fn"])
    if m:
        phenotype = m.group(1)
    else:
        phenotype = "unknown"
    return {phenotype: (data, cum_data, depth_1pc)}



""" MultiQC module to parse output from Bowtie 2 """


import logging
import re

from multiqc.modules.base_module import BaseMultiqcModule, ModuleNoSamplesFound
from multiqc.plots import bargraph

# Initialise the logger
log = logging.getLogger(__name__)


class MultiqcModule(BaseMultiqcModule):
    """Bowtie 2 module, parses stderr logs."""

    def __init__(self):
        # Initialise the parent object
        super(MultiqcModule, self).__init__(
            name="Bowtie 2 / HiSAT2",
            anchor="bowtie2",
            target="",
            info="""<a href="http://bowtie-bio.sourceforge.net/bowtie2/">Bowtie 2</a>
                and <a href="https://ccb.jhu.edu/software/hisat2/">HISAT2</a> are fast
                and memory-efficient tools for aligning sequencing reads against a reference genome.
                Unfortunately both tools have identical log output by default, so it is impossible
                to distiguish which tool was used.
                """,
            doi=["10.1038/nmeth.1923", "10.1038/nmeth.3317", "10.1038/s41587-019-0201-4"],
        )

        # Find and load any Bowtie 2 reports
        self.bowtie2_data = dict()
        self.num_se = 0
        self.num_pe = 0
        for f in self.find_log_files("bowtie2", filehandles=True):
            self.parse_bowtie2_logs(f)

        # Filter to strip out ignored sample names
        self.bowtie2_data = self.ignore_samples(self.bowtie2_data)

        if len(self.bowtie2_data) == 0:
            raise ModuleNoSamplesFound

        log.info(f"Found {len(self.bowtie2_data)} reports")

        # Superfluous function call to confirm that it is used in this module
        # Replace None with actual version if it is available
        self.add_software_version(None)

        # Write parsed report data to a file
        self.write_data_file(self.bowtie2_data, "multiqc_bowtie2")

        # Basic Stats Table
        # Report table is immutable, so just updating it works
        self.bowtie2_general_stats_table()

        # Alignment Rate Plot
        self.bowtie2_alignment_plot()

    def parse_bowtie2_logs(self, f):
        """
        Warning: This function may make you want to stab yourself.

        Parse logs from bowtie2. These miss several key bits of information
        such as input files, so we try to look for logs from other wrapper tools
        that may have logged this info. If not found, we default to using the filename.
        Note that concatenated logs only parse if we have the command printed in there.

        The bowtie log uses the same strings mulitple times in different contexts to mean
        different things, making parsing very messy. Handle with care.

        Example single-end output from bowtie2:
            Time loading reference: 00:00:08
            Time loading forward index: 00:00:16
            Time loading mirror index: 00:00:09
            [samopen] SAM header is present: 25 sequences.
            Multiseed full-index search: 00:58:04
            38377305 reads; of these:
              38377305 (100.00%) were unpaired; of these:
                2525577 (6.58%) aligned 0 times
                27593593 (71.90%) aligned exactly 1 time
                8258135 (21.52%) aligned >1 times
            93.42% overall alignment rate
            Time searching: 00:58:37
            Overall time: 00:58:37

        Example paired-end output from bowtie2:
            Time loading reference: 00:01:07
            Time loading forward index: 00:00:26
            Time loading mirror index: 00:00:09
            Multiseed full-index search: 01:32:55
            15066949 reads; of these:
              15066949 (100.00%) were paired; of these:
                516325 (3.43%) aligned concordantly 0 times
                11294617 (74.96%) aligned concordantly exactly 1 time
                3256007 (21.61%) aligned concordantly >1 times
                ----
                516325 pairs aligned concordantly 0 times; of these:
                  26692 (5.17%) aligned discordantly 1 time
                ----
                489633 pairs aligned 0 times concordantly or discordantly; of these:
                  979266 mates make up the pairs; of these:
                    592900 (60.55%) aligned 0 times
                    209206 (21.36%) aligned exactly 1 time
                    177160 (18.09%) aligned >1 times
            98.03% overall alignment rate
            Time searching: 01:34:37
            Overall time: 01:34:37
        """

        # Regexes
        regexes = {
            "unpaired": {
                "unpaired_aligned_none": r"(\d+) \([\d\.]+%\) aligned 0 times",
                "unpaired_aligned_one": r"(\d+) \([\d\.]+%\) aligned exactly 1 time",
                "unpaired_aligned_multi": r"(\d+) \([\d\.]+%\) aligned >1 times",
            },
            "paired": {
                "paired_aligned_none": r"(\d+) \([\d\.]+%\) aligned concordantly 0 times",
                "paired_aligned_one": r"(\d+) \([\d\.]+%\) aligned concordantly exactly 1 time",
                "paired_aligned_multi": r"(\d+) \([\d\.]+%\) aligned concordantly >1 times",
                "paired_aligned_discord_one": r"(\d+) \([\d\.]+%\) aligned discordantly 1 time",
                "paired_aligned_discord_multi": r"(\d+) \([\d\.]+%\) aligned discordantly >1 times",
                "paired_aligned_mate_one": r"(\d+) \([\d\.]+%\) aligned exactly 1 time",
                "paired_aligned_mate_multi": r"(\d+) \([\d\.]+%\) aligned >1 times",
                "paired_aligned_mate_none": r"(\d+) \([\d\.]+%\) aligned 0 times",
            },
        }

        # Go through log file line by line
        s_name = f["s_name"]
        parsed_data = {}

        for line in f["f"]:
            # Attempt in vain to find original bowtie2 command, logged by another program
            btcmd = re.search(r"bowtie2 .+ -[1U] ([^\s,]+)", line)
            if btcmd:
                s_name = self.clean_s_name(btcmd.group(1), f)
                log.debug(f"Found a bowtie2 command, updating sample name to '{s_name}'")

            # Total reads
            total = re.search(r"(\d+) reads; of these:", line)
            if total:
                parsed_data["total_reads"] = int(total.group(1))

            # Single end reads
            unpaired = re.search(r"(\d+) \([\d\.]+%\) were unpaired; of these:", line)
            if unpaired:
                parsed_data["unpaired_total"] = int(unpaired.group(1))
                self.num_se += 1

                # Do nested loop whilst we have this level of indentation
                line = f["f"].readline()
                while line.startswith("    "):
                    for k, r in regexes["unpaired"].items():
                        match = re.search(r, line)
                        if match:
                            parsed_data[k] = int(match.group(1))
                    line = f["f"].readline()

            # Paired end reads
            paired = re.search(r"(\d+) \([\d\.]+%\) were paired; of these:", line)
            if paired:
                parsed_data["paired_total"] = int(paired.group(1))
                self.num_pe += 1

                # Do nested loop whilst we have this level of indentation
                line = f["f"].readline()
                while line.startswith("    "):
                    for k, r in regexes["paired"].items():
                        match = re.search(r, line)
                        if match:
                            parsed_data[k] = int(match.group(1))
                    line = f["f"].readline()

            # Overall alignment rate
            overall = re.search(r"([\d\.]+)% overall alignment rate", line)
            if overall:
                parsed_data["overall_alignment_rate"] = float(overall.group(1))

                # End of log section
                # Save half 'pairs' of mate counts
                for k in ["paired_aligned_mate_multi", "paired_aligned_mate_none", "paired_aligned_mate_one"]:
                    if k in parsed_data:
                        parsed_data[f"{k}_halved"] = float(parsed_data[k]) / 2.0

                # HiSAT2 has PE data doesn't have the mate-specific stats.
                # To avoid missing unaligned counts in the barplot, fake the "_halved" key.
                # See https://github.com/MultiQC/MultiQC/issues/1230
                if "paired_aligned_mate_none_halved" not in parsed_data and "paired_aligned_none" in parsed_data:
                    parsed_data["paired_aligned_mate_none_halved"] = parsed_data["paired_aligned_none"]

                # Save parsed data
                if s_name in self.bowtie2_data:
                    log.debug(f"Duplicate sample name found! Overwriting: {s_name}")
                self.add_data_source(f, s_name)
                self.bowtie2_data[s_name] = parsed_data
                # Reset in case we find more in this log file
                s_name = f["s_name"]
                parsed_data = {}

    def bowtie2_general_stats_table(self):
        """Take the parsed stats from the Bowtie 2 report and add it to the
        basic stats table at the top of the report"""

        headers = {
            "overall_alignment_rate": {
                "title": "% Aligned",
                "description": "overall alignment rate",
                "max": 100,
                "min": 0,
                "suffix": "%",
                "scale": "YlGn",
            }
        }
        self.general_stats_addcols(self.bowtie2_data, headers)

    def bowtie2_alignment_plot(self):
        """Make the HighCharts HTML to plot the alignment rates"""

        half_warning = ""
        for s_name in self.bowtie2_data:
            if (
                "paired_aligned_mate_one_halved" in self.bowtie2_data[s_name]
                or "paired_aligned_mate_multi_halved" in self.bowtie2_data[s_name]
                or "paired_aligned_mate_none_halved" in self.bowtie2_data[s_name]
            ):
                half_warning = """
                <div class="alert alert-warning">Please note that single mate alignment counts are halved to tally with pair counts properly.</div>
                """
        description_text = "This plot shows the number of reads aligning to the reference in different ways."

        # Config for the plot
        config = {"ylab": "# Reads", "cpswitch_counts_label": "Number of Reads"}

        # Two plots, don't mix SE with PE
        if self.num_se > 0:
            sekeys = {
                "unpaired_aligned_one": {"color": "#20568f", "name": "SE mapped uniquely"},
                "unpaired_aligned_multi": {"color": "#f7a35c", "name": "SE multimapped"},
                "unpaired_aligned_none": {"color": "#981919", "name": "SE not aligned"},
            }
            config["id"] = "bowtie2_se_plot"
            config["title"] = "Bowtie 2: SE Alignment Scores"
            self.add_section(
                name="Single-end alignments",
                anchor="bowtie2-align-se",
                description=description_text,
                helptext="""
                There are 3 possible types of alignment:

                * **SE mapped uniquely**: Read has only one occurence in the reference genome.
                * **SE multimapped**: Read has multiple occurence.
                * **SE not aligned**: Read has no occurence.
                """,
                plot=bargraph.plot(self.bowtie2_data, sekeys, config),
            )

        if self.num_pe > 0:
            pekeys = {
                "paired_aligned_one": {"color": "#20568f", "name": "PE mapped uniquely"},
                "paired_aligned_discord_one": {"color": "#5c94ca", "name": "PE mapped discordantly uniquely"},
                "paired_aligned_mate_one_halved": {"color": "#95ceff", "name": "PE one mate mapped uniquely"},
                "paired_aligned_multi": {"color": "#f7a35c", "name": "PE multimapped"},
                "paired_aligned_discord_multi": {"color": "#dce333", "name": "PE discordantly multimapped"},
                "paired_aligned_mate_multi_halved": {"color": "#ffeb75", "name": "PE one mate multimapped"},
                "paired_aligned_mate_none_halved": {"color": "#981919", "name": "PE neither mate aligned"},
            }
            config["id"] = "bowtie2_pe_plot"
            config["title"] = "Bowtie 2: PE Alignment Scores"
            self.add_section(
                name="Paired-end alignments",
                anchor="bowtie2-align-pe",
                description=description_text + half_warning,
                helptext="""
                There are 6 possible types of alignment:

                * **PE mapped uniquely**: Pair has only one occurence in the reference genome.
                * **PE mapped discordantly uniquely**: Pair has only one occurence but not in proper pair.
                * **PE one mate mapped uniquely**: One read of a pair has one occurence.
                * **PE multimapped**: Pair has multiple occurence.
                * **PE one mate multimapped**: One read of a pair has multiple occurence.
                * **PE neither mate aligned**: Pair has no occurence.
                """,
                plot=bargraph.plot(self.bowtie2_data, pekeys, config),
            )



""" MultiQC functions to plot a table """

import logging

from multiqc.plots import table_object
from multiqc.utils import config
from multiqc.plots.plotly import table

logger = logging.getLogger(__name__)

letters = "abcdefghijklmnopqrstuvwxyz"

# Load the template so that we can access its configuration
# Do this lazily to mitigate import-spaghetti when running unit tests
_template_mod = None


def get_template_mod():
    global _template_mod
    if not _template_mod:
        _template_mod = config.avail_templates[config.template].load()
    return _template_mod


def plot(data, headers=None, pconfig=None):
    """Return HTML for a MultiQC table.
    :param data: 2D dict, first keys as sample names, then x:y data pairs
    :param headers: list of optional dicts with column config in key:value pairs.
    :return: HTML ready to be inserted into the page
    """
    # Make a datatable object
    dt = table_object.DataTable(data, headers, pconfig)

    # Collect unique sample names
    s_names = set()
    for d in dt.data:
        for s_name in d.keys():
            s_names.add(s_name)

    mod = get_template_mod()
    if "table" in mod.__dict__ and callable(mod.table):
        # noinspection PyBroadException
        try:
            return mod.table(dt, s_names, pconfig)
        except:  # noqa: E722
            if config.strict:
                # Crash quickly in the strict mode. This can be helpful for interactive
                # debugging of modules
                raise

    return table.plot(dt)



from multiqc.utils import config

from .plot_aqhist import plot_aqhist
from .plot_basic_hist import plot_basic_hist
from .plot_bhist import plot_bhist
from .plot_bqhist import plot_bqhist
from .plot_covhist import plot_covhist
from .plot_idhist import plot_idhist
from .plot_ihist import plot_ihist
from .plot_indelhist import plot_indelhist
from .plot_mhist import plot_mhist
from .plot_qahist import plot_qahist
from .plot_qchist import plot_qchist
from .plot_qhist import plot_qhist


section_order = [
    "stats",
    "covhist",
    "covstats",
    "bincov",
    "bqhist",
    "idhist",
    "indelhist",
    "mhist",
    "qahist",
    "qchist",
    "qhist",
    "aqhist",
    "ehist",
    "lhist",
    "ihist",
    "gchist",
    "bhist",
    "rpkm",
    "statsfile_machine",
    "statsfile",
]
file_types = {
    "stats": {
        "title": "BBDuk filtering statistics",
        "descr": "Proportion of reads that matched adapters/contaminants.",
        "help_text": "",
        "kvrows": ["Total", "Matched"],
        "kv_descriptions": {
            "Total": (
                "Total number of reads processed",
                {
                    "description": f"Aligned Reads ({config.read_count_desc})",
                    "shared_key": "read_count",
                    "modify": lambda x: x * config.read_count_multiplier,
                    "scale": "PuBu",
                    "hidden": True,
                },
            ),
            "Matched": (
                "Total number of reads matching adapters/contaminants",
                {
                    "description": f"Aligned Reads ({config.read_count_desc})",
                    "shared_key": "read_count",
                    "modify": lambda x: x * config.read_count_multiplier,
                    "scale": "Reds",
                    "hidden": True,
                },
            ),
            "Percent filtered": (
                "Proportion of reads filtered, matching adapters/contaminants",
                {"max": 100, "min": 0, "scale": "OrRd", "suffix": "%", "hidden": False},
            ),
        },
        "cols": {
            "Name": str,
            "Reads": int,
            "ReadsPct": lambda v: float(v.strip("%")),
        },
        "extracols": {
            "Bases": int,
            "BasesPct": float,
        },
        "plot_func": None,  # Plotting for 'stats' not implemented
        "plot_params": {},
    },
    "aqhist": {
        "title": "Read quality",
        "descr": "Histogram of average read qualities (`aqhist`). "
        "Plot shows the number of reads at each quality score.",
        "help_text": "",
        "cols": {"Quality": int, "count1": int, "fraction1": float, "count2": int, "fraction2": float},
        "plot_func": plot_aqhist,
        "plot_params": {
            "xPlotBands": [
                {"from": 28, "to": 100, "color": "#c3e6c3"},
                {"from": 20, "to": 28, "color": "#e6dcc3"},
                {"from": 0, "to": 20, "color": "#e6c3c3"},
            ],
            "yLog": True,
        },
    },
    # basecov
    #  - Coverage per base location.
    #  - too big to interpret here
    "bhist": {
        "title": "Base composition",
        "descr": "Base composition histogram by position (`bhist`). "
        "The plot shows the percentage of `G+C`, `A+T`, and `N` bases "
        "for each position in the reads.",
        "help_text": "Relative composition",
        "cols": {"Pos": int, "A": float, "C": float, "G": float, "T": float, "N": float},
        "plot_func": plot_bhist,
        "plot_params": {},
    },
    "bincov": {
        "title": "Binned coverage",
        "descr": "Binned coverage per location, one line per X bases (`bincov`).",
        "help_text": "",
        "kvrows": ["Mean", "STDev"],
        "cols": {"RefName": str, "Cov": float, "Pos": int, "RunningPos": int},
        "plot_func": plot_basic_hist,
        "plot_params": {},
        "not_implemented": "",
    },
    "bqhist": {
        "title": "Base quality",
        "descr": "Quality histogram designed for box plots (`bqhist`). "
        "Refer to original source files for complete boxplot data. "
        "Plot shows mean base quality for each read position. ",
        "help_text": "",
        "cols": {
            "BaseNum": int,
            "count_1": int,
            "min_1": int,
            "max_1": int,
            "mean_1": float,
            "Q1_1": int,
            "med_1": int,
            "Q3_1": int,
            "LW_1": int,
            "RW_1": int,
            "count_2": int,
            "min_2": int,
            "max_2": int,
            "mean_2": float,
            "Q1_2": int,
            "med_2": int,
            "Q3_2": int,
            "LW_2": int,
            "RW_2": int,
        },
        "plot_func": plot_bqhist,
        "plot_params": {},
    },
    "covhist": {
        "title": "Coverage histogram",
        "descr": "Histogram of number of occurrences of each coverage depth level (`covhist`). "
        "Note that lines have been smoothed to 400 points; "
        "higher resolution data might be available in the original data source. ",
        "help_text": "",
        "cols": {"Coverage": int, "numBases": int},
        "plot_func": plot_covhist,
        "plot_params": {
            "yLog": True,
        },
    },
    "covstats": {
        "title": "Coverage stats",
        "descr": "Per-scaffold coverage info (`covstats`).",
        "help_text": "",
        "cols": {"ID": str, "Avg_fold": float},
        "extracols": {
            "Length": int,
            "Ref_GC": float,
            "Covered_percent": float,
            "Covered_bases": int,
            "Plus_reads": int,
            "Minus_reads": int,
            "Median_fold": int,
            "Read_GC": float,
            "Std_Dev": float,
        },
        "plot_func": plot_basic_hist,
        "plot_params": {},
        "not_implemented": "",
    },
    "ehist": {
        "title": "Errors-per-read",
        "descr": "Errors-per-read histogram (`ehist`). ",
        "help_text": "",
        "cols": {"Errors": int, "Count": int},
        "plot_func": plot_basic_hist,
        "plot_params": {"xlab": "Errors", "ylab": "# Reads"},
    },
    "gchist": {
        "title": "GC content",
        "descr": "Read GC content histogram (`gchist`).",
        "help_text": "",
        "kvrows": ["Mean", "Median", "Mode", "STDev"],
        "kv_descriptions": {
            "Mean": ("Average GC content", {}),
            "Median": ("Median GC content", {}),
            "Mode": ("The most commonly occuring value of the GC content distribution", {}),
            "STDev": ("Standard deviation of average GC content", {}),
        },
        "cols": {"GC": float, "Count": int},
        "plot_func": plot_basic_hist,
        "plot_params": {"xlab": "Proportion GC", "ylab": "# Reads"},
    },
    "idhist": {
        "title": "Identity histogram",
        "descr": "Histogram of read count versus percent base pair identity " "of aligned reads (`idhist`).",
        "help_text": "",
        "kvrows": [
            "Mean_reads",
            "Mean_bases",
            "Median_reads",
            "Median_bases",
            "Mode_reads",
            "Mode_bases",
            "STDev_reads",
            "STDev_bases",
        ],
        "kv_descriptions": {
            "Mean_reads": ("Average percent identity of aligned reads", {}),
            "Mean_bases": ("Average percent identity of aligned bases", {}),
            "Median_reads": ("Median percent identity of aligned reads", {}),
            "Median_bases": ("Median percent identity of aligned bases", {}),
            "Mode_reads": (
                "The most commonly occuring average value amongst aligned reads (i.e. the mode of the distribution)",
                {},
            ),
            "Mode_bases": (
                "The most commonly occuring average value amongst aligned bases (i.e. the mode of the distribution)",
                {},
            ),
            "STDev_reads": (
                "The standard deviation of the average percent identity distribution for aligned reads",
                {},
            ),
            "STDev_bases": (
                "The standard deviation of the average percent identity distribution for aligned bases",
                {},
            ),
        },
        "cols": {"Identity": float, "Reads": int, "Bases": int},
        "plot_func": plot_idhist,
        "plot_params": {},
    },
    "ihist": {
        "title": "Insert sizes",
        "descr": "Histogram of computed insert sizes, for paired reads (`ihist`). "
        "Plotted data has been cut off at 99% to prevent long tails; "
        "Complete data available in original source files.",
        "help_text": "The insert size is the length of the sequence between the "
        "sequencing adapters, which for most common insert sizes is "
        "longer than the sum of both read pairs. "
        "In some cases, the insert size is shorter than the length "
        "of the two read pairs combined, resulting in an insert size "
        "shorter than the sum of the length of the reads pairs.",
        "kvrows": ["Mean", "Median", "STDev", "PercentOfPairs"],
        "kv_descriptions": {
            "Mean": ("Average insert length", {}),
            "Median": ("Median insert length", {}),
            "STDev": ("Standard deviation of insert size length distribution", {}),
            "PercentOfPairs": ("", {}),
        },
        "cols": {"InsertSize": int, "Count": int},
        "plot_func": plot_ihist,
        "plot_params": {},
    },
    "indelhist": {
        "title": "Indel lengths",
        "descr": "Indel length histogram (`indelhist`). "
        "The plots show the number of observed insertions and deletions, "
        "for each insertion and deletion length.",
        "help_text": "",
        "cols": {"Length": int, "Deletions": int, "Insertions": int},
        "plot_func": plot_indelhist,
        "plot_params": {},
    },
    "lhist": {
        "title": "Read lengths",
        "descr": "Read length histogram (`lhist`).",
        "help_text": "",
        "cols": {"Length": int, "Count": int},
        "plot_func": plot_basic_hist,
        "plot_params": {"xlab": "Read length (base pairs)", "ylab": "# Reads"},
    },
    "mhist": {
        "title": "Match, substitution, deletion, and insertion rates",
        "descr": "Histogram of match, substitution, deleletion, " "and insertion rates by read location (`mhist`).",
        "help_text": "",
        "cols": {
            "BaseNum": int,
            "Match1": float,
            "Sub1": float,
            "Del1": float,
            "Ins1": float,
            "N1": float,
            "Other1": float,
            "Match2": float,
            "Sub2": float,
            "Del2": float,
            "Ins2": float,
            "N2": float,
            "Other2": float,
        },
        "plot_func": plot_mhist,
        "plot_params": {},
    },
    "qahist": {
        "title": "Quality accuracy",
        "descr": "Base quality accuracy histogram of error rates versus quality score (`qahist`). "
        "The plots show the observed count of each type of alignment by base quality score.",
        "help_text": "",
        "kvrows": ["Deviation", "DeviationSub"],
        "kv_descriptions": {
            "Deviation": ("", {}),
            "DeviationSub": ("", {}),
        },
        "cols": {
            "Quality": int,
            "Match": int,
            "Sub": int,
            "Ins": int,
            "Del": int,
            "TrueQuality": float,
            "TrueQualitySub": float,
        },
        "plot_func": plot_qahist,
        "plot_params": {},
    },
    "qchist": {
        "title": "Count of bases with each quality value",
        "descr": "Histogram of base qualities (`qchist`). "
        "Plot shows the number of bases at each quality score. Zero counts are shown as `0.1` due to log axis.",
        "help_text": "",
        "cols": {"Quality": int, "count1": int, "fraction1": float},
        "plot_func": plot_qchist,
        "plot_params": {
            "xPlotBands": [
                {"from": 30, "to": 100, "color": "#c3e6c3"},
                {"from": 20, "to": 30, "color": "#e6dcc3"},
                {"from": 0, "to": 20, "color": "#e6c3c3"},
            ],
            "yLog": True,
            "xlab": "Phred Score",
            "ylab": "Counts",
        },
    },
    "qhist": {
        "title": "Sequence Quality Histograms",
        "descr": "Quality histogram by position (`qhist`). "
        "The plots show the average quality for each position in the reads, "
        "using the linear values, logarithmically scaled values, and the "
        "actual measured qualities based on the alignments.",
        "help_text": "",
        "cols": {
            "BaseNum": int,
            "Read1_linear": float,
            "Read1_log": float,
            "Read1_measured": float,
            "Read2_linear": float,
            "Read2_log": float,
            "Read2_measured": float,
        },
        "plot_func": plot_qhist,
        "plot_params": {},
    },
    "rpkm": {
        "title": "RPKM/FPKM",
        "descr": "Per-scaffold RPKM/FPKM counts (`rpkm`).",
        "help_text": "",
        "kvrows": ["File", "Reads", "Mapped", "RefSequences"],
        "kv_descriptions": {
            "File": ("", {}),
            "Reads": ("", {}),
            "Mapped": ("", {}),
            "RefSequences": ("", {}),
        },
        "cols": {
            "Name": str,
            "Length": int,
            "Bases": int,
            "Coverage": float,
            "Reads": int,
            "RPKM": float,
            "Frags": int,
            "FPKM": float,
        },
        "plot_func": plot_basic_hist,
        "plot_params": {},
        "not_implemented": "",
    },
    "statsfile_machine": {
        "title": "General stats",
        "descr": "General Stats (`statsfile_machine`).",
        "help_text": "",
        "cols": [],
        "plot_func": plot_basic_hist,
        "plot_params": {},
    },
    "statsfile": {
        "title": "General stats",
        "descr": "General Stats (`statsfile`).",
        "help_text": "",
        "cols": [],
        "plot_params": {},
        "plot_func": plot_basic_hist,
        "not_implemented": "",
    },
}

statsfile_machine_keys = [
    "Reads_Used",
    "Bases_Used",
    "Reads/sec",
    "kBases/sec",
    "R1_Mapped_Percent",
    "R1_Unambiguous_Percent",
    "R1_Mapped_Reads",
    "R1_Unambiguous_Reads",
    "Mated_Pairs",
    "Bad_Pairs",
    "R1_Rescued",
    "Avg_Insert_Size",
    "R1_Perfect_Best_Site",
    "R1_Semiperfect_Site",
    "R1_Ambiguous_Mapping",
    "R1_Low_Quality_Discards",
    "R1_Match_Rate",
    "R1_Error_Rate",
    "R1_Sub_Rate",
    "R1_Del_Rate",
    "R1_Ins_Rate",
    "R1_N_Rate",
    "R1_Match_Count",
    "R1_Error_Count",
    "R1_Sub_Count",
    "R1_Del_Count",
    "R1_Ins_Count",
    "R1_N_Count",
    "R2_Mapped_Percent",
    "R2_Unambiguous_Percent",
    "R2_Mapped_Reads",
    "R2_Unambiguous_Reads",
    "R2_Rescued",
    "R2_Perfect_Best_Site",
    "R2_Semiperfect_Site",
    "R2_Ambiguous_Mapping",
    "R2_Low_Quality_Discards",
    "R2_Match_Rate",
    "R2_Error_Rate",
    "R2_Sub_Rate",
    "R2_Del_Rate",
    "R2_Ins_Rate",
    "R2_N_Rate",
    "R2_Match_Count",
    "R2_Error_Count",
    "R2_Sub_Count",
    "R2_Del_Count",
    "R2_Ins_Count",
    "R2_N_Count",
    "R2_Mapped_Percent",
]



#!/usr/bin/env python

""" Super Special-Case MultiQC module to produce report section on software versions """


import logging
from textwrap import dedent

from multiqc.modules.base_module import BaseMultiqcModule
from multiqc.utils import config as mqc_config
from multiqc.utils import report as mqc_report
from multiqc.utils import util_functions

# Initialise the logger
log = logging.getLogger(__name__)


class MultiqcModule(BaseMultiqcModule):
    def __init__(self):
        # Initialise the parent object
        super(MultiqcModule, self).__init__(
            name="Software Versions",
            anchor="multiqc_software_versions",
            info="lists versions of software tools extracted from file contents.",
        )

        self.report_software_versions()
        self.write_software_versions_data_file()

    def report_software_versions(self):
        """Create section listing software versions."""
        content = self._make_versions_html(mqc_report.software_versions)
        self.add_section(name=None, content=content)

    @staticmethod
    def _make_versions_html(versions):
        """Generate a tabular HTML output of all versions."""
        table_id = mqc_report.save_htmlid("mqc_versions_table")

        # Check if the Group column is identical to Software column
        groups_rows = []
        software_rows = []
        for group, tmp_versions in sorted(versions.items()):
            for tool, _ in sorted(tmp_versions.items()):
                groups_rows.append(group)
                software_rows.append(tool)
        ignore_groups = groups_rows == software_rows

        # Based on: https://github.com/nf-core/rnaseq/blob/3bec2331cac2b5ff88a1dc71a21fab6529b57a0f/modules/nf-core/custom/dumpsoftwareversions/templates/dumpsoftwareversions.py#L12
        header_rows = ["<th>Software</th>", "<th>Version</th>"]
        if not ignore_groups:
            header_rows.insert(0, f"<th>{mqc_config.versions_table_group_header}</th>")
        html = [
            dedent(
                f"""\
                <button type="button" class="mqc_table_copy_btn btn btn-default btn-sm" data-clipboard-target="table#{table_id}">
                    <span class="glyphicon glyphicon-copy"></span> Copy table
                </button>
                <table class="table mqc_versions_table" id="{table_id}">
                    <thead>
                        <tr>{''.join(header_rows)}</tr>
                    </thead>
                """
            )
        ]
        for group, tmp_versions in sorted(versions.items()):
            html.append("<tbody>")
            for i, (tool, versions) in enumerate(sorted(tmp_versions.items())):
                rows = [
                    f"<td>{tool}</td>",
                    f"<td><samp>{', '.join(list(map(str, versions)))}</samp></td>",
                ]
                if not ignore_groups:
                    rows.insert(0, f"<td>{group if (i == 0) else ''}</td>")
                html.append(f"<tr>{''.join(rows)}</tr>")
            html.append("</tbody>")
        html.append("</table>")
        return "\n".join(html)

    @staticmethod
    def write_software_versions_data_file():
        """Write software versions to a file for downstream use."""
        # Get rid of the default dicts and Version objects
        flat_software_versions = {
            group: {software: list(map(str, software_versions)) for software, software_versions in versions.items()}
            for group, versions in mqc_report.software_versions.items()
        }
        # TSV only allows 2 levels of nesting.
        if mqc_config.data_format == "tsv":
            flat_software_versions = {
                group: {software: ", ".join(software_versions) for software, software_versions in versions.items()}
                for group, versions in flat_software_versions.items()
            }
        # Write to a file for downstream use
        util_functions.write_data_file(flat_software_versions, "multiqc_software_versions")



#!/usr/bin/env python

""" MultiQC module to parse output from sourmash """

import logging

from multiqc.modules.base_module import BaseMultiqcModule, ModuleNoSamplesFound

from .compare import CompareMixin
from .gather import GatherMixin

# Initialise the logger
log = logging.getLogger(__name__)


class MultiqcModule(BaseMultiqcModule, CompareMixin, GatherMixin):
    def __init__(self):
        super(MultiqcModule, self).__init__(
            name="Sourmash",
            anchor="sourmash",
            href="https://github.com/sourmash-bio/sourmash",
            info="quickly searches, compares, and analyzes genomic and metagenomic data sets.",
            doi="10.21105/joss.00027",
        )

        n = dict()
        n["compare"] = self.parse_compare()
        if n["compare"] > 0:
            log.info(f"Found {n['compare']} compare results")

        n["gather"] = self.parse_gather()
        if n["gather"] > 0:
            log.info(f"Found {n['gather']} gather results")

        if sum(n.values()) == 0:
            raise ModuleNoSamplesFound



#!/usr/bin/env python
"""
Helper functions to manipulate colours and colour scales
"""
import hashlib

# Default logger will be replaced by caller
import logging
import re

import numpy as np
import spectra

from multiqc.utils import config, report

logger = logging.getLogger(__name__)


class mqc_colour_scale(object):
    """Class to hold a colour scheme."""

    # ColorBrewer colours, taken from Chroma.js source code
    # https://github.com/gka/chroma.js
    ###    Copyright (c) 2002 Cynthia Brewer, Mark Harrower, and The
    ###    Pennsylvania State University.
    ###
    ###    Licensed under the Apache License, Version 2.0 (the "License");
    ###    you may not use this file except in compliance with the License.
    ###    You may obtain a copy of the License at
    ###    http://www.apache.org/licenses/LICENSE-2.0
    ###
    ###    Unless required by applicable law or agreed to in writing, software distributed
    ###    under the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
    ###    CONDITIONS OF ANY KIND, either express or implied. See the License for the
    ###    specific language governing permissions and limitations under the License.
    COLORBREWER_SCALES = {
        # sequential
        "OrRd": ["#fff7ec", "#fee8c8", "#fdd49e", "#fdbb84", "#fc8d59", "#ef6548", "#d7301f", "#b30000", "#7f0000"],
        "PuBu": ["#fff7fb", "#ece7f2", "#d0d1e6", "#a6bddb", "#74a9cf", "#3690c0", "#0570b0", "#045a8d", "#023858"],
        "BuPu": ["#f7fcfd", "#e0ecf4", "#bfd3e6", "#9ebcda", "#8c96c6", "#8c6bb1", "#88419d", "#810f7c", "#4d004b"],
        "Oranges": [
            "#fff5eb",
            "#fee6ce",
            "#fdd0a2",
            "#fdae6b",
            "#fd8d3c",
            "#f16913",
            "#d94801",
            "#a63603",
            "#7f2704",
        ],
        "BuGn": ["#f7fcfd", "#e5f5f9", "#ccece6", "#99d8c9", "#66c2a4", "#41ae76", "#238b45", "#006d2c", "#00441b"],
        "YlOrBr": [
            "#ffffe5",
            "#fff7bc",
            "#fee391",
            "#fec44f",
            "#fe9929",
            "#ec7014",
            "#cc4c02",
            "#993404",
            "#662506",
        ],
        "YlGn": ["#ffffe5", "#f7fcb9", "#d9f0a3", "#addd8e", "#78c679", "#41ab5d", "#238443", "#006837", "#004529"],
        "Reds": ["#fff5f0", "#fee0d2", "#fcbba1", "#fc9272", "#fb6a4a", "#ef3b2c", "#cb181d", "#a50f15", "#67000d"],
        "RdPu": ["#fff7f3", "#fde0dd", "#fcc5c0", "#fa9fb5", "#f768a1", "#dd3497", "#ae017e", "#7a0177", "#49006a"],
        "Greens": [
            "#f7fcf5",
            "#e5f5e0",
            "#c7e9c0",
            "#a1d99b",
            "#74c476",
            "#41ab5d",
            "#238b45",
            "#006d2c",
            "#00441b",
        ],
        "YlGnBu": [
            "#ffffd9",
            "#edf8b1",
            "#c7e9b4",
            "#7fcdbb",
            "#41b6c4",
            "#1d91c0",
            "#225ea8",
            "#253494",
            "#081d58",
        ],
        "Purples": [
            "#fcfbfd",
            "#efedf5",
            "#dadaeb",
            "#bcbddc",
            "#9e9ac8",
            "#807dba",
            "#6a51a3",
            "#54278f",
            "#3f007d",
        ],
        "GnBu": ["#f7fcf0", "#e0f3db", "#ccebc5", "#a8ddb5", "#7bccc4", "#4eb3d3", "#2b8cbe", "#0868ac", "#084081"],
        "Greys": [
            "#ffffff",
            "#f0f0f0",
            "#d9d9d9",
            "#bdbdbd",
            "#969696",
            "#737373",
            "#525252",
            "#252525",
            "#000000",
        ],
        "YlOrRd": [
            "#ffffcc",
            "#ffeda0",
            "#fed976",
            "#feb24c",
            "#fd8d3c",
            "#fc4e2a",
            "#e31a1c",
            "#bd0026",
            "#800026",
        ],
        "PuRd": ["#f7f4f9", "#e7e1ef", "#d4b9da", "#c994c7", "#df65b0", "#e7298a", "#ce1256", "#980043", "#67001f"],
        "Blues": [
            "#f7fbff",
            "#deebf7",
            "#c6dbef",
            "#9ecae1",
            "#6baed6",
            "#4292c6",
            "#2171b5",
            "#08519c",
            "#08306b",
        ],
        "PuBuGn": [
            "#fff7fb",
            "#ece2f0",
            "#d0d1e6",
            "#a6bddb",
            "#67a9cf",
            "#3690c0",
            "#02818a",
            "#016c59",
            "#014636",
        ],
        # diverging
        "Spectral": [
            "#9e0142",
            "#d53e4f",
            "#f46d43",
            "#fdae61",
            "#fee08b",
            "#ffffbf",
            "#e6f598",
            "#abdda4",
            "#66c2a5",
            "#3288bd",
            "#5e4fa2",
        ],
        "RdYlGn": [
            "#a50026",
            "#d73027",
            "#f46d43",
            "#fdae61",
            "#fee08b",
            "#ffffbf",
            "#d9ef8b",
            "#a6d96a",
            "#66bd63",
            "#1a9850",
            "#006837",
        ],
        "RdBu": [
            "#67001f",
            "#b2182b",
            "#d6604d",
            "#f4a582",
            "#fddbc7",
            "#f7f7f7",
            "#d1e5f0",
            "#92c5de",
            "#4393c3",
            "#2166ac",
            "#053061",
        ],
        "PiYG": [
            "#8e0152",
            "#c51b7d",
            "#de77ae",
            "#f1b6da",
            "#fde0ef",
            "#f7f7f7",
            "#e6f5d0",
            "#b8e186",
            "#7fbc41",
            "#4d9221",
            "#276419",
        ],
        "PRGn": [
            "#40004b",
            "#762a83",
            "#9970ab",
            "#c2a5cf",
            "#e7d4e8",
            "#f7f7f7",
            "#d9f0d3",
            "#a6dba0",
            "#5aae61",
            "#1b7837",
            "#00441b",
        ],
        "RdYlBu": [
            "#a50026",
            "#d73027",
            "#f46d43",
            "#fdae61",
            "#fee090",
            "#ffffbf",
            "#e0f3f8",
            "#abd9e9",
            "#74add1",
            "#4575b4",
            "#313695",
        ],
        "BrBG": [
            "#543005",
            "#8c510a",
            "#bf812d",
            "#dfc27d",
            "#f6e8c3",
            "#f5f5f5",
            "#c7eae5",
            "#80cdc1",
            "#35978f",
            "#01665e",
            "#003c30",
        ],
        "RdGy": [
            "#67001f",
            "#b2182b",
            "#d6604d",
            "#f4a582",
            "#fddbc7",
            "#ffffff",
            "#e0e0e0",
            "#bababa",
            "#878787",
            "#4d4d4d",
            "#1a1a1a",
        ],
        "PuOr": [
            "#7f3b08",
            "#b35806",
            "#e08214",
            "#fdb863",
            "#fee0b6",
            "#f7f7f7",
            "#d8daeb",
            "#b2abd2",
            "#8073ac",
            "#542788",
            "#2d004b",
        ],
        # qualitative
        "Set2": ["#66c2a5", "#fc8d62", "#8da0cb", "#e78ac3", "#a6d854", "#ffd92f", "#e5c494", "#b3b3b3"],
        "Accent": ["#7fc97f", "#beaed4", "#fdc086", "#ffff99", "#386cb0", "#f0027f", "#bf5b17", "#666666"],
        "Set1": ["#e41a1c", "#377eb8", "#4daf4a", "#984ea3", "#ff7f00", "#ffff33", "#a65628", "#f781bf", "#999999"],
        "Set3": [
            "#8dd3c7",
            "#ffffb3",
            "#bebada",
            "#fb8072",
            "#80b1d3",
            "#fdb462",
            "#b3de69",
            "#fccde5",
            "#d9d9d9",
            "#bc80bd",
            "#ccebc5",
            "#ffed6f",
        ],
        "Dark2": ["#1b9e77", "#d95f02", "#7570b3", "#e7298a", "#66a61e", "#e6ab02", "#a6761d", "#666666"],
        "Paired": [
            "#a6cee3",
            "#1f78b4",
            "#b2df8a",
            "#33a02c",
            "#fb9a99",
            "#e31a1c",
            "#fdbf6f",
            "#ff7f00",
            "#cab2d6",
            "#6a3d9a",
            "#ffff99",
            "#b15928",
        ],
        "Pastel2": ["#b3e2cd", "#fdcdac", "#cbd5e8", "#f4cae4", "#e6f5c9", "#fff2ae", "#f1e2cc", "#cccccc"],
        "Pastel1": [
            "#fbb4ae",
            "#b3cde3",
            "#ccebc5",
            "#decbe4",
            "#fed9a6",
            "#ffffcc",
            "#e5d8bd",
            "#fddaec",
            "#f2f2f2",
        ],
        # Originally from Highcharts
        "plot_defaults": [
            "#7cb5ec",
            "#434348",
            "#90ed7d",
            "#f7a35c",
            "#8085e9",
            "#f15c80",
            "#e4d354",
            "#2b908f",
            "#f45b5b",
            "#91e8e1",
        ],
    }

    def __init__(self, name="GnBu", minval=0, maxval=100, id=None):
        """Initialise class with a colour scale"""

        self.name = name
        self.id = id
        self.colours = self.get_colours(name)

        # Sanity checks
        minval = re.sub(r"[^0-9\.\-e]", "", str(minval))
        maxval = re.sub(r"[^0-9\.\-e]", "", str(maxval))
        if minval == "":
            minval = 0
        if maxval == "":
            maxval = 100
        if float(minval) == float(maxval):
            self.minval = float(minval)
            self.maxval = float(minval) + 1.0
        elif float(minval) > float(maxval):
            self.minval = float(maxval)
            self.maxval = float(minval)
        else:
            self.minval = float(minval)
            self.maxval = float(maxval)

    def get_colour(self, val, colformat="hex", lighten=0.3, source=None):
        """Given a value, return a colour within the colour scale"""

        if val is None:
            return ""

        # Ported from the original JavaScript for continuity
        # Seems to work better than adjusting brightness / saturation / luminosity
        def rgb_converter(x):
            return max(0, min(1, 1 + ((x - 1) * lighten)))

        try:
            if self.name in mqc_colour_scale.qualitative_scales and isinstance(val, float):
                if config.strict:
                    sequential_scales = [
                        s for s in mqc_colour_scale.COLORBREWER_SCALES if s not in mqc_colour_scale.qualitative_scales
                    ]
                    source = f"{source}: " if source else ""
                    errmsg = (
                        f"{source}A categorical scale '{self.name}' is used for float values ({val}). "
                        f"Consider using one of the sequential scales instead: {', '.join(sequential_scales)}"
                    )
                    logger.error(errmsg)
                    report.lint_errors.append(errmsg)
            elif self.name in mqc_colour_scale.qualitative_scales:
                if not isinstance(val, int):
                    # When we have non-numeric values (e.g. Male/Female, Yes/No, chromosome names, etc.), and a qualitative
                    # scale (Set1, Set3, etc.), we don't want to attempt to parse numbers, otherwise we might end up with all
                    # values assigned with the same color. But instead we will get a hash from a string to hope to assign
                    # a unique color for each possible enumeration value.
                    val = deterministic_hash(val)
                thecolour = spectra.html(self.colours[val % len(self.colours)])
                thecolour = spectra.rgb(*[rgb_converter(v) for v in thecolour.rgb])
                return thecolour.hexcode

            # When there is only 1 color in scale, spectra.scale() will crash with DivisionByZero
            elif len(self.colours) == 1:
                thecolour = spectra.html(self.colours[0])
                thecolour = spectra.rgb(*[rgb_converter(v) for v in thecolour.rgb])
                return thecolour.hexcode

            else:
                # Sanity checks
                val_stripped = re.sub(r"[^0-9\.\-e]", "", str(val))
                val_float: float
                if val_stripped == "":
                    val_float = self.minval
                else:
                    try:
                        val_float = float(val_stripped)
                    except ValueError:
                        return ""
                    val_float = max(val_float, self.minval)
                    val_float = min(val_float, self.maxval)

                domain_nums = list(np.linspace(self.minval, self.maxval, len(self.colours)))
                my_scale = spectra.scale(self.colours).domain(domain_nums)

                # Lighten colours
                thecolour = spectra.rgb(*[rgb_converter(v) for v in my_scale(val_float).rgb])

                return thecolour.hexcode

        except Exception as e:
            # Shouldn't crash all of MultiQC just for colours
            logger.warning(f"{self.id + ': ' if self.id else ''}Error getting colour: {e}")
            return ""

    def get_colours(self, name="GnBu"):
        """Function to get a colour scale by name
        Input: Name of colour scale (suffix with -rev for reversed)
               Defaults to 'GnBu' if scale not found.
        Returns: List of hex colours
        """
        if name.startswith("#"):
            return [name]

        if name in mqc_colour_scale.html_colors:
            return [name]

        # Detect reverse colour scales
        reverse = False
        if str(name).endswith("-rev"):
            reverse = True
            name = name[:-4]

        # Default colour scale
        if name not in mqc_colour_scale.COLORBREWER_SCALES:
            errmsg = f"{self.id + ': ' if self.id else ''}Colour scale {name} not found - defaulting to GnBu"
            if config.strict:
                logger.error(errmsg)
                report.lint_errors.append(errmsg)
            else:
                logger.debug(errmsg)
            name = "GnBu"

        # Return colours
        if reverse:
            return list(reversed(mqc_colour_scale.COLORBREWER_SCALES[name]))
        else:
            return mqc_colour_scale.COLORBREWER_SCALES[name]

    qualitative_scales = ["Set2", "Accent", "Set1", "Set3", "Dark2", "Paired", "Pastel2", "Pastel1", "plot_defaults"]

    html_colors = {
        "black": "#000000",
        "silver": "#C0C0C0",
        "gray": "#808080",
        "grey": "#808080",
        "white": "#FFFFFF",
        "maroon": "#800000",
        "red": "#FF0000",
        "purple": "#800080",
        "fuchsia": "#FF00FF",
        "green": "#008000",
        "lime": "#00FF00",
        "olive": "#808000",
        "yellow": "#FFFF00",
        "navy": "#000080",
        "blue": "#0000FF",
        "teal": "#008080",
        "aqua": "#00FFFF",
        "darkblue": "#00008B",
        "mediumblue": "#0000CD",
        "darkgreen": "#006400",
        "darkcyan": "#008B8B",
        "deepskyblue": "#00BFFF",
        "darkturquoise": "#00CED1",
        "mediumspringgreen": "#00FA9A",
        "springgreen": "#00FF7F",
        "cyan": "#00FFFF",
        "midnightblue": "#191970",
        "dodgerblue": "#1E90FF",
        "lightseagreen": "#20B2AA",
        "forestgreen": "#228B22",
        "seagreen": "#2E8B57",
        "darkslategray": "#2F4F4F",
        "darkslategrey": "#2F4F4F",
        "limegreen": "#32CD32",
        "mediumseagreen": "#3CB371",
        "turquoise": "#40E0D0",
        "royalblue": "#4169E1",
        "steelblue": "#4682B4",
        "darkslateblue": "#483D8B",
        "mediumturquoise": "#48D1CC",
        "indigo": "#4B0082",
        "darkolivegreen": "#556B2F",
        "cadetblue": "#5F9EA0",
        "cornflowerblue": "#6495ED",
        "rebeccapurple": "#663399",
        "mediumaquamarine": "#66CDAA",
        "dimgray": "#696969",
        "dimgrey": "#696969",
        "slateblue": "#6A5ACD",
        "olivedrab": "#6B8E23",
        "slategray": "#708090",
        "slategrey": "#708090",
        "lightslategray": "#778899",
        "lightslategrey": "#778899",
        "mediumslateblue": "#7B68EE",
        "lawngreen": "#7CFC00",
        "chartreuse": "#7FFF00",
        "aquamarine": "#7FFFD4",
        "skyblue": "#87CEEB",
        "lightskyblue": "#87CEFA",
        "blueviolet": "#8A2BE2",
        "darkred": "#8B0000",
        "darkmagenta": "#8B008B",
        "saddlebrown": "#8B4513",
        "darkseagreen": "#8FBC8F",
        "lightgreen": "#90EE90",
        "mediumpurple": "#9370DB",
        "darkviolet": "#9400D3",
        "palegreen": "#98FB98",
        "darkorchid": "#9932CC",
        "yellowgreen": "#9ACD32",
        "sienna": "#A0522D",
        "brown": "#A52A2A",
        "darkgray": "#A9A9A9",
        "darkgrey": "#A9A9A9",
        "lightblue": "#ADD8E6",
        "greenyellow": "#ADFF2F",
        "paleturquoise": "#AFEEEE",
        "lightsteelblue": "#B0C4DE",
        "powderblue": "#B0E0E6",
        "firebrick": "#B22222",
        "darkgoldenrod": "#B8860B",
        "mediumorchid": "#BA55D3",
        "rosybrown": "#BC8F8F",
        "darkkhaki": "#BDB76B",
        "mediumvioletred": "#C71585",
        "indianred": "#CD5C5C",
        "peru": "#CD853F",
        "chocolate": "#D2691E",
        "tan": "#D2B48C",
        "lightgray": "#D3D3D3",
        "lightgrey": "#D3D3D3",
        "thistle": "#D8BFD8",
        "orchid": "#DA70D6",
        "goldenrod": "#DAA520",
        "palevioletred": "#DB7093",
        "crimson": "#DC143C",
        "gainsboro": "#DCDCDC",
        "plum": "#DDA0DD",
        "burlywood": "#DEB887",
        "lightcyan": "#E0FFFF",
        "lavender": "#E6E6FA",
        "darksalmon": "#E9967A",
        "violet": "#EE82EE",
        "palegoldenrod": "#EEE8AA",
        "lightcoral": "#F08080",
        "khaki": "#F0E68C",
        "aliceblue": "#F0F8FF",
        "honeydew": "#F0FFF0",
        "azure": "#F0FFFF",
        "sandybrown": "#F4A460",
        "wheat": "#F5DEB3",
        "beige": "#F5F5DC",
        "whitesmoke": "#F5F5F5",
        "mintcream": "#F5FFFA",
        "ghostwhite": "#F8F8FF",
        "salmon": "#FA8072",
        "antiquewhite": "#FAEBD7",
        "linen": "#FAF0E6",
        "lightgoldenrodyellow": "#FAFAD2",
        "oldlace": "#FDF5E6",
        "magenta": "#FF00FF",
        "deeppink": "#FF1493",
        "orangered": "#FF4500",
        "tomato": "#FF6347",
        "hotpink": "#FF69B4",
        "coral": "#FF7F50",
        "darkorange": "#FF8C00",
        "lightsalmon": "#FFA07A",
        "orange": "#FFA500",
        "lightpink": "#FFB6C1",
        "pink": "#FFC0CB",
        "gold": "#FFD700",
        "peachpuff": "#FFDAB9",
        "navajowhite": "#FFDEAD",
        "moccasin": "#FFE4B5",
        "bisque": "#FFE4C4",
        "mistyrose": "#FFE4E1",
        "blanchedalmond": "#FFEBCD",
        "papayawhip": "#FFEFD5",
        "lavenderblush": "#FFF0F5",
        "seashell": "#FFF5EE",
        "cornsilk": "#FFF8DC",
        "lemonchiffon": "#FFFACD",
        "floralwhite": "#FFFAF0",
        "snow": "#FFFAFA",
        "lightyellow": "#FFFFE0",
        "ivory": "#FFFFF0",
    }


def deterministic_hash(x):
    """
    Deterministic hash function for strings. This is useful for assigning a unique color
    to each possible value of a categorical variable.

    Unlike the built-in hash function, this function always returns the same value for
    the same input string (see https://docs.python.org/3/using/cmdline.html#cmdoption-R)
    """
    return int(hashlib.sha1(x.encode("utf-8")).hexdigest(), 16) % (10**8)



import logging

from multiqc import config
from multiqc.modules.base_module import BaseMultiqcModule, ModuleNoSamplesFound
from multiqc.plots import beeswarm

log = logging.getLogger(__name__)


class MultiqcModule(BaseMultiqcModule):
    def __init__(self):
        # Initialise the parent object
        super(MultiqcModule, self).__init__(
            name="PRINSEQ++",
            anchor="prinseqplusplus",
            href="https://github.com/Adrian-Cantu/PRINSEQ-plus-plus",
            info="PRINSEQ++ is a C++ implementation of the prinseq-lite.pl program. It can be used to filter, reformat or trim genomic and metagenomic sequence data.",
            doi="10.7287/peerj.preprints.27553v1",
        )

        # Find and load reports
        self.prinseqplusplus_data = dict()

        # Find all files for prinseqplusplus
        for f in self.find_log_files("prinseqplusplus", filehandles=True):
            self.parse_logs(f)

        self.prinseqplusplus_data = self.ignore_samples(self.prinseqplusplus_data)

        if len(self.prinseqplusplus_data) == 0:
            raise ModuleNoSamplesFound

        log.info(f"Found {len(self.prinseqplusplus_data)} reports")

        # Superfluous function call to confirm that it is used in this module
        # Replace None with actual version if it is available
        self.add_software_version(None)

        # Write data to file
        self.write_data_file(self.prinseqplusplus_data, "prinseqplusplus")

        self.prinseqplusplus_general_stats()
        self.prinseqplusplus_beeswarm_plot()

    def parse_logs(self, f):
        """Parsing Logs."""
        s_name = f["s_name"]
        if self.prinseqplusplus_data.get(s_name) is not None:
            log.warn(f"Duplicate sample name found! Overwriting: {s_name}")

        self.prinseqplusplus_data[s_name] = {}
        self.add_data_source(f, s_name=s_name)

        for line in f["f"]:
            # Find line after loading reads, and remove suffixes for sample name
            if "-min_len" in line:
                self.prinseqplusplus_data[s_name]["min_len"] = int(line.lstrip().rstrip().split(" ")[0])
            elif "-max_len" in line:
                self.prinseqplusplus_data[s_name]["max_len"] = int(line.lstrip().rstrip().split(" ")[0])
            elif "-min_cg" in line:  ## This seems to be a typo in the log file of PRINSEQ++, should be gc
                self.prinseqplusplus_data[s_name]["min_gc"] = int(line.lstrip().rstrip().split(" ")[0])
            elif "-max_cg" in line:
                self.prinseqplusplus_data[s_name]["max_gc"] = int(line.lstrip().rstrip().split(" ")[0])
            elif "-min_qual_score" in line:
                self.prinseqplusplus_data[s_name]["min_qual_score"] = int(line.lstrip().rstrip().split(" ")[0])
            elif "-min_qual_mean" in line:
                self.prinseqplusplus_data[s_name]["min_qual_mean"] = int(line.lstrip().rstrip().split(" ")[0])
            elif "-ns_max_n" in line:
                self.prinseqplusplus_data[s_name]["ns_max_n"] = int(line.lstrip().rstrip().split(" ")[0])
            elif "-noiupac" in line:
                self.prinseqplusplus_data[s_name]["noiupac"] = int(line.lstrip().rstrip().split(" ")[0])
            elif "-derep" in line:
                self.prinseqplusplus_data[s_name]["derep"] = int(line.lstrip().rstrip().split(" ")[0])
            elif "-lc_entropy" in line:
                self.prinseqplusplus_data[s_name]["lc_entropy"] = int(line.lstrip().rstrip().split(" ")[0])
            elif "-lc_dust" in line:
                self.prinseqplusplus_data[s_name]["lc_dust"] = int(line.lstrip().rstrip().split(" ")[0])

    def prinseqplusplus_general_stats(self):
        """PRINSEQ++ General Stats Table"""
        data = {}
        for s_name, d in self.prinseqplusplus_data.items():
            data[s_name] = {"prinseqplusplus_total": sum(d.values())}

        self.general_stats_addcols(
            data,
            {
                "prinseqplusplus_total": {
                    "title": f"Filtered Reads ({config.read_count_prefix})",
                    "description": f"Sum of filtered reads ({config.read_count_desc})",
                    "scale": "Oranges",
                    "shared_key": "read_count",
                    "modify": lambda x: x * config.read_count_multiplier,
                }
            },
        )

    def prinseqplusplus_beeswarm_plot(self):
        """Beeswarm plot of all possible filtering results"""
        # This would be nicer as a stacked-bar plot, but as we don't have
        # the total read count it doesn't really make sense.

        reads = {
            "min": 0,
            "modify": lambda x: float(x) * config.read_count_multiplier,
            "suffix": config.read_count_prefix,
            "shared_key": "read_count",
        }
        headers = {
            "min_len": dict(reads, title="min_len"),
            "max_len": dict(reads, title="max_len"),
            "min_gc": dict(reads, title="min_gc"),
            "max_gc": dict(reads, title="max_gc"),
            "lc_entropy": dict(reads, title="lc_entropy"),
            "min_qual_score": dict(reads, title="min_qual_score"),
            "min_qual_mean": dict(reads, title="min_qual_mean"),
            "ns_max_n": dict(reads, title="ns_max_n"),
            "noiupac": dict(reads, title="noiupac"),
            "derep": dict(reads, title="derep"),
            "lc_dust": dict(reads, title="lc_dust"),
        }

        self.add_section(
            name="Filtered Reads",
            anchor="prinseqplusplus-filtered-reads",
            description="Shows the number of reads removed by the various PRINSEQ++ filter options",
            plot=beeswarm.plot(
                self.prinseqplusplus_data,
                headers,
                {
                    "id": "prinseplusplus-filtered-reads-beeswarm",
                    "title": "PRINSEQ++: Filtered Reads",
                },
            ),
        )



""" MultiQC submodule to parse output from Picard MarkDuplicates """

import logging
import math
from collections import defaultdict

from multiqc import config
from multiqc.modules.picard import util
from multiqc.plots import bargraph

# Initialise the logger
log = logging.getLogger(__name__)


def parse_reports(module, sp_key="picard/markdups"):
    """
    Find Picard MarkDuplicates reports and parse their data.
    Note that this function is also used by the biobambam2 module, that's why
    the parameter.
    """

    data_by_sample = dict()

    # Get custom config value
    try:
        merge_multiple_libraries = config.picard_config["markdups_merge_multiple_libraries"]
    except (AttributeError, KeyError):
        merge_multiple_libraries = True

    # Function to save results at end of table
    def save_table_results(s_name, keys, parsed_data, recompute_merged_metrics):
        # No data
        if len(keys) == 0 or len(parsed_data) == 0:
            return

        # User has requested each library is kept separate
        # Update the sample name to append the library name
        if not merge_multiple_libraries and len(parsed_data) > 0:
            s_name = f"{s_name} - {parsed_data['LIBRARY']}"

        # Skip - No reads
        try:
            if parsed_data["READ_PAIRS_EXAMINED"] == 0 and parsed_data["UNPAIRED_READS_EXAMINED"] == 0:
                log.warning(f"Skipping MarkDuplicates sample '{s_name}' as log contained no reads")
                return
        # Skip - Missing essential fields
        except KeyError:
            log.warning(f"Skipping MarkDuplicates sample '{s_name}' as missing essential fields")
            return

        # Recompute PERCENT_DUPLICATION and ESTIMATED_LIBRARY_SIZE
        if recompute_merged_metrics:
            parsed_data["PERCENT_DUPLICATION"] = calculate_percentage_duplication(parsed_data)
            parsed_data["ESTIMATED_LIBRARY_SIZE"] = estimate_library_size(parsed_data)

        # Save the data
        if s_name in data_by_sample:
            log.debug(f"Duplicate sample name found in {f['fn']}! Overwriting: {s_name}")
        data_by_sample[s_name] = parsed_data
        module.add_data_source(f, s_name, section="DuplicationMetrics")

        # End of metrics table - reset for next sample
        if len(vals) < 6:
            return True

        # On to the next library if not merging
        else:
            return False

    # Go through logs and find Metrics
    for f in module.find_log_files(sp_key, filehandles=True):
        s_name = f["s_name"]
        parsed_lists = defaultdict(list)
        keys = None
        in_stats_block = False
        recompute_merged_metrics = False

        for line in f["f"]:
            maybe_s_name = util.extract_sample_name(
                module,
                line,
                f,
                picard_tool="MarkDuplicates",
                sentieon_algo="Dedup",
            )
            if maybe_s_name:
                s_name = maybe_s_name

            if s_name is None:
                continue

            if util.is_line_right_before_table(line, picard_class="DuplicationMetric", sentieon_algo="Dedup"):
                keys = f["f"].readline().strip("\n").split("\t")
                if s_name in data_by_sample:
                    log.debug(f"Duplicate sample name found in {f['fn']}! Overwriting: {s_name}")
                in_stats_block = True

            # Currently parsing the METRICS table
            elif in_stats_block:
                vals = line.rstrip("\n").split("\t")

                # End of the METRICS table, or multiple libraries, and we're not merging them
                if len(vals) < 6 or (not merge_multiple_libraries and len(parsed_lists) > 0):
                    parsed_data = {k: parsed_list[0] for k, parsed_list in parsed_lists.items()}
                    if save_table_results(s_name, keys, parsed_data, recompute_merged_metrics):
                        # Reset for next file if returned True
                        s_name = None
                        parsed_lists = defaultdict(list)
                        keys = None
                        in_stats_block = False
                        recompute_merged_metrics = False
                    continue

                # Parse the column values
                assert len(vals) == len(keys), (keys, vals, f)
                for k, v in zip(keys, vals):
                    # More than one library present and merging stats
                    if k in parsed_lists:
                        recompute_merged_metrics = True

                    v = v.strip()
                    try:
                        v = float(v)
                    except ValueError:
                        pass
                    parsed_lists[k].append(v)

        parsed_data = {}
        for k in parsed_lists:
            # Sometimes a numerical column will an empty string, so converting "" to 0.0
            if all(isinstance(x, float) or x == "" for x in parsed_lists[k]):
                parsed_data[k] = sum(0.0 if x == "" else x for x in parsed_lists[k])
            else:
                parsed_data[k] = "/".join(str(x) for x in parsed_lists[k])

        # Files with no extra lines after last library
        if in_stats_block:
            save_table_results(s_name, keys, parsed_data, recompute_merged_metrics)

    # Filter to strip out ignored sample names
    data_by_sample = module.ignore_samples(data_by_sample)
    if len(data_by_sample) == 0:
        return 0

    # Superfluous function call to confirm that it is used in this module
    # Replace None with actual version if it is available
    module.add_software_version(None)

    # Write parsed data to a file
    module.write_data_file(data_by_sample, f"multiqc_{module.anchor}_dups")

    # Add to general stats table
    headers = {
        "PERCENT_DUPLICATION": {
            "title": "Duplication",
            "description": "Mark Duplicates - Percent Duplication",
            "max": 100,
            "min": 0,
            "suffix": "%",
            "scale": "OrRd",
            "modify": lambda x: util.multiply_hundred(x),
        }
    }
    module.general_stats_addcols(data_by_sample, headers, namespace="Mark Duplicates")

    # Make the bar plot and add to the MarkDuplicates section
    #
    # The table in the Picard metrics file contains some columns referring
    # read pairs and some referring to single reads.
    for s_name, metr in data_by_sample.items():
        metr["READS_IN_DUPLICATE_PAIRS"] = 2.0 * metr["READ_PAIR_DUPLICATES"]
        metr["READS_IN_UNIQUE_PAIRS"] = 2.0 * (metr["READ_PAIRS_EXAMINED"] - metr["READ_PAIR_DUPLICATES"])
        metr["READS_IN_UNIQUE_UNPAIRED"] = metr["UNPAIRED_READS_EXAMINED"] - metr["UNPAIRED_READ_DUPLICATES"]
        metr["READS_IN_DUPLICATE_PAIRS_OPTICAL"] = 2.0 * metr["READ_PAIR_OPTICAL_DUPLICATES"]
        metr["READS_IN_DUPLICATE_PAIRS_NONOPTICAL"] = (
            metr["READS_IN_DUPLICATE_PAIRS"] - metr["READS_IN_DUPLICATE_PAIRS_OPTICAL"]
        )
        metr["READS_IN_DUPLICATE_UNPAIRED"] = metr["UNPAIRED_READ_DUPLICATES"]
        metr["READS_UNMAPPED"] = metr["UNMAPPED_READS"]

    keys = dict()
    keys_r = [
        "READS_IN_UNIQUE_PAIRS",
        "READS_IN_UNIQUE_UNPAIRED",
        "READS_IN_DUPLICATE_PAIRS_OPTICAL",
        "READS_IN_DUPLICATE_PAIRS_NONOPTICAL",
        "READS_IN_DUPLICATE_UNPAIRED",
        "READS_UNMAPPED",
    ]
    for k in keys_r:
        keys[k] = {"name": k.replace("READS_", "").replace("IN_", "").replace("_", " ").title()}

    # Config for the plot
    pconfig = {
        "id": f"{module.anchor}_deduplication",
        "title": f"{module.name}: Deduplication Stats",
        "ylab": "# Reads",
        "cpswitch_counts_label": "Number of Reads",
        "cpswitch_c_active": False,
    }

    module.add_section(
        name="Mark Duplicates",
        anchor=f"{module.anchor}-markduplicates",
        description="Number of reads, categorised by duplication state. **Pair counts "
        "are doubled** - see help text for details.",
        helptext="""
        The table in the Picard metrics file contains some columns referring
        read pairs and some referring to single reads.

        To make the numbers in this plot sum correctly, values referring to pairs are 
        doubled
        according to the scheme below:

        * `READS_IN_DUPLICATE_PAIRS = 2 * READ_PAIR_DUPLICATES`
        * `READS_IN_UNIQUE_PAIRS = 2 * (READ_PAIRS_EXAMINED - READ_PAIR_DUPLICATES)`
        * `READS_IN_UNIQUE_UNPAIRED = UNPAIRED_READS_EXAMINED - 
        UNPAIRED_READ_DUPLICATES`
        * `READS_IN_DUPLICATE_PAIRS_OPTICAL = 2 * READ_PAIR_OPTICAL_DUPLICATES`
        * `READS_IN_DUPLICATE_PAIRS_NONOPTICAL = READS_IN_DUPLICATE_PAIRS - 
        READS_IN_DUPLICATE_PAIRS_OPTICAL`
        * `READS_IN_DUPLICATE_UNPAIRED = UNPAIRED_READ_DUPLICATES`
        * `READS_UNMAPPED = UNMAPPED_READS`
        """,
        plot=bargraph.plot(data_by_sample, keys, pconfig),
    )

    # Return the number of detected samples to the parent module
    return len(data_by_sample)


def calculate_percentage_duplication(d):
    """
    Picard calculation to compute percentage duplication

    Taken & translated from the Picard codebase:
    https://github.com/broadinstitute/picard/blob
    /78ea24466d9bcab93db89f22e6a6bf64d0ad7782/src/main/java/picard/sam
    /DuplicationMetrics.java#L106-L110
    """
    try:
        dups = d["UNPAIRED_READ_DUPLICATES"] + d["READ_PAIR_DUPLICATES"] * 2
        examined = d["UNPAIRED_READS_EXAMINED"] + d["READ_PAIRS_EXAMINED"] * 2
        if examined == 0:
            return 0
        return float(dups) / float(examined)
    except KeyError:
        return 0


def estimate_library_size(d):
    """
    Picard calculation to estimate library size

    Taken & translated from the Picard codebase:
    https://github.com/broadinstitute/picard/blob
    /78ea24466d9bcab93db89f22e6a6bf64d0ad7782/src/main/java/picard/sam
    /DuplicationMetrics.java#L153-L164

    Note: Optical duplicates are contained in duplicates and therefore do not enter
    the calculation here.
    See also the computation of READ_PAIR_NOT_OPTICAL_DUPLICATES.

     * Estimates the size of a library based on the number of paired end molecules
     observed
     * and the number of unique pairs observed.
     * <p>
     * Based on the Lander-Waterman equation that states:
     * C/X = 1 - exp( -N/X )
     * where
     * X = number of distinct molecules in library
     * N = number of read pairs
     * C = number of distinct fragments observed in read pairs
    """

    try:
        readPairs = d["READ_PAIRS_EXAMINED"] - d["READ_PAIR_OPTICAL_DUPLICATES"]
        uniqueReadPairs = d["READ_PAIRS_EXAMINED"] - d["READ_PAIR_DUPLICATES"]
    except KeyError:
        return None

    readPairDuplicates = readPairs - uniqueReadPairs

    if readPairs > 0 and readPairDuplicates > 0:
        m = 1.0
        M = 100.0

        if uniqueReadPairs >= readPairs or f(m * uniqueReadPairs, uniqueReadPairs, readPairs) < 0:
            logging.warning("Picard recalculation of ESTIMATED_LIBRARY_SIZE skipped - metrics " "look wrong")
            return None

        # find value of M, large enough to act as other side for bisection method
        while f(M * uniqueReadPairs, uniqueReadPairs, readPairs) > 0:
            M *= 10.0

        # use bisection method (no more than 40 times) to find solution
        for i in range(40):
            r = (m + M) / 2.0
            u = f(r * uniqueReadPairs, uniqueReadPairs, readPairs)
            if u == 0:
                break
            elif u > 0:
                m = r
            elif u < 0:
                M = r

        return uniqueReadPairs * (m + M) / 2.0
    else:
        return None


def f(x, c, n):
    """
    Picard calculation used when estimating library size

    Taken & translated from the Picard codebase:
    https://github.com/broadinstitute/picard/blob
    /78ea24466d9bcab93db89f22e6a6bf64d0ad7782/src/main/java/picard/sam
    /DuplicationMetrics.java#L172-L177

    * Method that is used in the computation of estimated library size.
    """
    return c / x - 1 + math.exp(-n / x)



""" MultiQC submodule to parse output from deepTools bamPEFragmentSize for summary table """

import logging

from multiqc.plots import linegraph, table

# Initialise the logger
log = logging.getLogger(__name__)


class bamPEFragmentSizeTableMixin:
    def parse_bamPEFragmentSize(self):
        """Find bamPEFragmentSize output. Supports the --table option"""
        self.deeptools_bamPEFragmentSize = dict()
        for f in self.find_log_files("deeptools/bamPEFragmentSizeTable"):
            parsed_data = self.parseBamPEFile(f)
            for k, v in parsed_data.items():
                if k in self.deeptools_bamPEFragmentSize:
                    log.warning(f"Replacing duplicate sample {k}.")
                self.deeptools_bamPEFragmentSize[k] = v

            if len(parsed_data) > 0:
                self.add_data_source(f, section="bamPEFragmentSize")

        self.deeptools_bamPEFragmentSize = self.ignore_samples(self.deeptools_bamPEFragmentSize)

        if len(self.deeptools_bamPEFragmentSize) == 0:
            return 0

        # Superfluous function call to confirm that it is used in this module
        # Replace None with actual version if it is available
        self.add_software_version(None)

        # Write data to file
        self.write_data_file(self.deeptools_bamPEFragmentSize, "deeptools_frag_size_table")

        headersSE = {
            "Reads Sampled": {
                "title": "# Sampled",
                "description": "Number of reads sampled",
                "format": "{:,.0f}",
            },
            "Read Len. Min.": {
                "title": "Min",
                "description": "Minimum read length",
                "format": "{:,.0f}",
                "shared_key": "read_length",
            },
            "Read Len. 1st. Qu.": {
                "title": "1st Quartile",
                "description": "1st quartile read length",
                "format": "{:,.0f}",
                "shared_key": "read_length",
            },
            "Read Len. Mean": {
                "title": "Mean",
                "description": "Mean read length",
                "shared_key": "read_length",
            },
            "Read Len. Median": {
                "title": "Median",
                "description": "Median read length",
                "format": "{:,.0f}",
                "shared_key": "read_length",
            },
            "Read Len. 3rd Qu.": {
                "title": "3rd Quartile",
                "description": "3rd quartile read length",
                "format": "{:,.0f}",
                "shared_key": "read_length",
            },
            "Read Len. Max": {
                "title": "Max",
                "description": "Maximum read length",
                "format": "{:,.0f}",
                "shared_key": "read_length",
            },
            "Read Len. Std.": {
                "title": "Std. Dev.",
                "description": "read length standard deviation",
                "shared_key": "read_length",
            },
            "Read Med. Abs. Dev.": {
                "title": "MAD",
                "description": "read length median absolute deviation",
                "shared_key": "read_length",
            },
        }
        config = {
            "namespace": "deepTools bamPEFragmentSize",
            "id": "deeptools_readlengths_table",
            "title": "deepTools: Read length metrics",
        }
        self.add_section(
            name="Read length metrics",
            anchor="deeptools_readlengths",
            plot=table.plot(self.deeptools_bamPEFragmentSize, headersSE, config),
        )

        headersPE = {
            "Frag. Sampled": {
                "title": "# Sampled",
                "description": "Number of fragments sampled",
                "format": "{:,.0f}",
            },
            "Frag. Len. Min.": {
                "title": "Min",
                "description": "Minimum fragment length",
                "format": "{:,.0f}",
                "shared_key": "frag_length",
            },
            "Frag. Len. 1st. Qu.": {
                "title": "1st Quartile",
                "description": "1st quartile fragment length",
                "format": "{:,.0f}",
                "shared_key": "frag_length",
            },
            "Frag. Len. Mean": {
                "title": "Mean",
                "description": "Mean fragment length",
                "format": "{:,.0f}",
                "shared_key": "frag_length",
            },
            "Frag. Len. Median": {
                "title": "Median",
                "description": "Median fragment length",
                "format": "{:,.0f}",
                "shared_key": "frag_length",
            },
            "Frag. Len. 3rd Qu.": {
                "title": "3rd Quartile",
                "description": "3rd quartile fragment length",
                "format": "{:,.0f}",
                "shared_key": "frag_length",
            },
            "Frag. Len. Max": {
                "title": "Max",
                "description": "Maximum fragment length",
                "format": "{:,.0f}",
                "shared_key": "frag_length",
            },
            "Frag. Len. Std.": {
                "title": "Std. Dev.",
                "description": "Fragment length standard deviation",
                "shared_key": "frag_length",
            },
            "Frag. Med. Abs. Dev.": {
                "title": "MAD",
                "description": "Fragment length median absolute deviation",
                "shared_key": "frag_length",
            },
        }

        # Are there any PE datasets?
        PE = False
        for k, v in self.deeptools_bamPEFragmentSize.items():
            if "Frag. Len. Min." in v:
                PE = True
                break
        if PE:
            config = {
                "namespace": "deepTools bamPEFragmentSize",
                "id": "deeptools_fragmentlengths_table",
                "title": "deepTools: Fragment length metrics",
            }
            self.add_section(
                name="Fragment length metrics",
                anchor="deeptools_fragmentlengths",
                plot=table.plot(self.deeptools_bamPEFragmentSize, headersPE, config),
            )

        # Read length plot
        config = {
            "data_labels": [
                {
                    "name": "Read length distribution",
                    "title": "Read length distribution",
                    "ylab": "Read length (bases)",
                },
                {
                    "name": "Fragment length distribution",
                    "title": "Fragment length distribution",
                    "ylab": "Fragment length (bases)",
                },
            ],
            "id": "deeptools_readlengthsPlot",
            "title": "deepTools: Read/Fragment length distribution",
            "namespace": "deepTools bamPEFragmentSize",
            "ylab": "Read length (bases)",
            "xlab": "Percentile",
        }
        SE = dict()
        PE = dict()
        for k, v in self.deeptools_bamPEFragmentSize.items():
            SE[k] = {
                0: v["Read Len. Min."],
                10: v["Read Len. 10%"],
                20: v["Read Len. 20%"],
                25: v["Read Len. 1st. Qu."],
                30: v["Read Len. 30%"],
                40: v["Read Len. 40%"],
                50: v["Read Len. Median"],
                60: v["Read Len. 60%"],
                70: v["Read Len. 70%"],
                75: v["Read Len. 3rd Qu."],
                80: v["Read Len. 80%"],
                90: v["Read Len. 90%"],
                99: v["Read Len. 99%"],
                100: v["Read Len. Max"],
            }
            if "Frag. Len. Min." not in v:
                continue
            PE[k] = {
                0: v["Frag. Len. Min."],
                10: v["Frag. Len. 10%"],
                20: v["Frag. Len. 20%"],
                25: v["Frag. Len. 1st. Qu."],
                30: v["Frag. Len. 30%"],
                40: v["Frag. Len. 40%"],
                50: v["Frag. Len. Median"],
                60: v["Frag. Len. 60%"],
                70: v["Frag. Len. 70%"],
                75: v["Frag. Len. 3rd Qu."],
                80: v["Frag. Len. 80%"],
                90: v["Frag. Len. 90%"],
                99: v["Frag. Len. 99%"],
                100: v["Frag. Len. Max"],
            }
        self.add_section(
            name="Read/fragment length distribution",
            anchor="deeptools_fragmentlengths_dist",
            plot=linegraph.plot([SE, PE], config),
        )

        return len(self.deeptools_bamPEFragmentSize)

    def parseBamPEFile(self, f):
        d = {}
        headers = None
        for line in f["f"].splitlines():
            cols = line.rstrip().split("\t")
            if headers is None:
                headers = cols
            else:
                s_name = None
                for idx, h in enumerate(headers):
                    if idx == 0:
                        s_name = self.clean_s_name(cols[0], f)
                        if s_name in d:
                            log.debug(f"Replacing duplicate sample {s_name}.")
                        d[s_name] = dict()
                    else:
                        if idx < 19 and cols[1] == "0":
                            # Don't store fragment metrics for SE datasets, they're just 0.
                            continue
                        try:
                            # Most values are ac
                            d[s_name][h] = self._int(cols[idx])
                        except ValueError:
                            d[s_name][h] = float(cols[idx])

        return d



import csv
import logging
import os
import xml.etree.ElementTree as ET
from collections import defaultdict
import functools
from itertools import islice
from typing import Tuple, Dict, Optional

from multiqc import config
from multiqc.modules.base_module import BaseMultiqcModule, ModuleNoSamplesFound
from multiqc.plots import bargraph, table

log = logging.getLogger(__name__)


class MultiqcModule(BaseMultiqcModule):
    def __init__(self):
        # Initialise the parent object
        super(MultiqcModule, self).__init__(
            name="BCL Convert",
            anchor="bclconvert",
            href="https://support.illumina.com/sequencing/sequencing_software/bcl-convert.html",
            info="can be used to both demultiplex data and convert BCL files to FASTQ file formats for downstream analysis.",
            # Can't find a DOI // doi=
        )

        # Set up and collate bclconvert run and demux files
        demuxes, qmetrics, multiple_sequencing_runs, last_run_id = self._collate_log_files()

        # variables to store reads for undetermined read recalculation
        self.per_lane_undetermined_reads = dict()
        self.total_reads_in_lane_per_file = dict()
        bclconvert_data = dict()
        for demux in demuxes.values():
            self.parse_demux_data(demux, bclconvert_data, len(demuxes))
        for qmetric in qmetrics.values():
            self.parse_qmetrics_data(bclconvert_data, qmetric)

        if len(demuxes) == 0:
            raise ModuleNoSamplesFound
        elif len(demuxes) > 1 and not multiple_sequencing_runs:
            log.warning("Found multiple runs from the same sequencer output")
            self.intro += """
                <div class="alert alert-warning">
                    <strong>Warning:</strong> Detected multiple bclconvert runs from the same sequencer output.
                    Runs were merged and undetermined stats were recalculated.
                </div>
            """
            self._recalculate_undetermined(bclconvert_data, last_run_id)
        elif multiple_sequencing_runs:
            # If we have data from multiple sequencing runs, the recalculation in
            # _recalculate_undetermined(last_run_id) wont work. In this case we
            # suppress/hide the info.
            log.warning("Found multiple sequencer runs")
            self.intro += """
                <div class="alert alert-warning">
                    <strong>Warning:</strong> Detected multiple sequencer runs.
                    Sample stats were merged.
                </div>
            """
            self.per_lane_undetermined_reads = None

        create_undetermined_barplots = (
            getattr(config, "bclconvert", {}).get("create_undetermined_barcode_barplots", False) or len(demuxes) == 1
        )
        if create_undetermined_barplots:
            self._parse_top_unknown_barcodes(bclconvert_data, last_run_id)

        # Collect counts by lane and sample
        (
            bclconvert_by_lane,
            bclconvert_by_sample,
            counts_by_sample_by_lane,
            source_files,
        ) = self._split_data_by_lane_and_sample(bclconvert_data)

        # Filter to strip out ignored sample names
        bclconvert_by_lane = self.ignore_samples(bclconvert_by_lane)
        bclconvert_by_sample = self.ignore_samples(bclconvert_by_sample)
        counts_by_sample_by_lane = self.ignore_samples(counts_by_sample_by_lane)

        # Return with Warning if no files are found
        if len(bclconvert_by_lane) == 0 and len(bclconvert_by_sample) == 0:
            raise ModuleNoSamplesFound
        log.info(f"{len(bclconvert_by_lane)} lanes and {len(bclconvert_by_sample)} samples found")

        # Print source files
        for s in source_files.keys():
            self.add_data_source(
                s_name=s,
                source=",".join(list(set(source_files[s]))),
                module="bclconvert",
                section="bclconvert-bysample",
            )

        # Calculate mean quality scores
        for sample_id, sample in bclconvert_by_sample.items():
            if sample["yield"] > 0:
                sample["mean_quality"] = sample["_quality_score_sum"] / sample["yield"]
            del sample["_quality_score_sum"]

        for lane_id, lane in bclconvert_by_lane.items():
            if lane["yield"] > 0:
                lane["mean_quality"] = lane["_quality_score_sum"] / lane["yield"]
            del lane["_quality_score_sum"]

        self.write_data_file(bclconvert_by_lane, "multiqc_bclconvert_bylane")
        self.write_data_file(bclconvert_by_sample, "multiqc_bclconvert_bysample")

        # Superfluous function call to confirm that it is used in this module
        # Replace None with actual version if it is available
        self.add_software_version(None)

        # Add sections for summary stats per flow cell
        self.add_section(
            name="Sample Statistics",
            anchor="bclconvert-samplestats",
            description="Statistics about each sample for each flowcell",
            plot=self.sample_stats_table(bclconvert_data, bclconvert_by_sample),
        )

        self.add_section(
            name="Lane Statistics",
            anchor="bclconvert-lanestats",
            description="Statistics about each lane for each flowcell",
            plot=self.lane_stats_table(bclconvert_by_lane),
        )

        # Add section for counts by lane
        cats = {
            "perfect": {"name": "Perfect Index Reads"},
            "imperfect": {"name": "Mismatched Index Reads"},
            "undetermined": {"name": "Undetermined Reads"},
        }
        extra = ""
        if len(demuxes) > 1 and not multiple_sequencing_runs:
            extra = """
                <div class="alert alert-warning">
                    <strong>Warning:</strong> Found multiple runs from the same sequencer output.
                    Runs were merged and <em>Undetermined Reads</em> were recalculated.
                </div>
            """
        elif multiple_sequencing_runs:
            extra = """
                <div class="alert alert-warning">
                    <strong>Warning:</strong> Found multiple sequencer runs.
                    <em>Undetermined Reads</em> cannot be recalculated for multiple sequencing runs and are not shown.
                </div>
            """
        self.add_section(
            name="Clusters by lane",
            anchor="bclconvert-bylane",
            description="Number of reads per lane (with number of perfect index reads)." + extra,
            helptext="""Perfect index reads are those that do not have a single mismatch.
                All samples of a lane are combined. Undetermined reads are treated as a third category.""",
            plot=bargraph.plot(
                self.get_bar_data_from_counts(bclconvert_data, bclconvert_by_lane, last_run_id),
                cats,
                {
                    "id": "bclconvert_lane_counts",
                    "title": "bclconvert: Clusters by lane",
                    "ylab": "Number of clusters",
                    "hide_zero_cats": False,
                },
            ),
        )

        self.add_section(
            name="Clusters by sample",
            anchor="bclconvert-bysample",
            description="Number of reads per sample.",
            helptext="""Perfect index reads are those that do not have a single mismatch.
                All samples are aggregated across lanes combined. Undetermined reads are ignored.
                Undetermined reads are treated as a separate sample.""",
            plot=bargraph.plot(
                [
                    self.get_bar_data_from_counts(bclconvert_data, bclconvert_by_sample, last_run_id),
                    counts_by_sample_by_lane,
                ],
                [cats, sorted(bclconvert_by_lane.keys())],
                {
                    "id": "bclconvert_sample_counts",
                    "title": "bclconvert: Clusters by sample",
                    "hide_zero_cats": False,
                    "ylab": "Number of clusters",
                    "data_labels": ["Index mismatches", "Counts per lane"],
                },
            ),
        )

        # Add section with undetermined barcodes
        if create_undetermined_barplots:
            self.add_section(
                name="Undetermined barcodes by lane",
                anchor="undetermine_by_lane",
                description="Undetermined barcodes by lanes",
                plot=bargraph.plot(
                    self.get_bar_data_from_undetermined(bclconvert_by_lane),
                    None,
                    {
                        "id": "bclconvert_undetermined",
                        "title": "bclconvert: Undetermined barcodes by lane",
                        "ylab": "Count",
                        "tt_percentages": False,
                        "use_legend": True,
                        "tt_suffix": "reads",
                    },
                ),
            )

    @staticmethod
    @functools.lru_cache
    def _get_genome_size() -> Optional[int]:
        gs = getattr(config, "bclconvert", {}).get("genome_size")
        if gs:
            try:
                gs = int(float(gs))
            except ValueError:
                presets = {"hg19_genome": 2897310462, "hg38_genome": 3049315783, "mm10_genome": 2652783500}
                if gs in presets:
                    gs = presets[gs]
                else:
                    log.warning(
                        f"Genome '{gs}' is unknown to MultiQC, please specify size explicitly or choose one of the "
                        f"following: {', '.join(presets.keys())}."
                    )
                    gs = None
        log.debug(f"Determined genome size as: {gs}")
        return gs

    @staticmethod
    def _reads_dictionary():
        return {
            "clusters": 0,
            "yield": 0,
            "_calculated_yield": 0,  # re-calculated yield from demux stats where it is not provided explicitly and is calculated from # Reads and Read Length
            "perfect_index_reads": 0,
            "one_mismatch_index_reads": 0,
            "basesQ30": 0,
            "depth": 0,
            "_quality_score_sum": 0,  # used to re-calculate mean_quality
            "_calculated_quality_score_sum": 0,  # used to re-calculate mean_quality from demux stats where Yield is not provided explicitly and is calculated from # Reads and Read Length
        }

    @staticmethod
    def _is_single_end_reads(root):
        ncount = 0
        # we have single-end reads if we have exactly one IsIndexedRead=N element in our RunInfo.xml
        for element in root.findall("./Run/Reads/Read"):
            if element.get("IsIndexedRead") == "N":
                ncount += 1
        if ncount == 1:
            return True
        return False

    @staticmethod
    def _get_r2_length(root):
        for element in root.findall("./Run/Reads/Read"):
            if element.get("Number") == "3" and element.get("IsIndexedRead") == "N":
                return element.get("NumCycles")  # single-index paired-end data
            if element.get("Number") == "4" and element.get("IsIndexedRead") == "N":
                return element.get("NumCycles")

        log.error("Could not figure out read 2 length from RunInfo.xml")
        raise ModuleNoSamplesFound

    def _parse_single_runinfo_file(self, runinfo_file):
        # Get run id and cluster length from RunInfo.xml
        try:
            root = ET.fromstring(runinfo_file["f"])
            run_id = root.find("Run").get("Id")
            read_length_r1 = root.find("./Run/Reads/Read[1]").get(
                "NumCycles"
            )  # ET indexes first element at 1, so here we're getting the first NumCycles
        except:  # noqa: E722
            log.error(f"Could not parse RunInfo.xml to get RunID and read length in '{runinfo_file['root']}'")
            raise ModuleNoSamplesFound
        if self._is_single_end_reads(root):
            cluster_length = int(read_length_r1)
        else:
            read_length_r2 = self._get_r2_length(root)
            cluster_length = int(read_length_r1) + int(read_length_r2)
        return {"run_id": run_id, "cluster_length": cluster_length}

    def _collate_log_files(self):
        # This function returns a list of self.find_log_files('bclconvert/demux') dicts,
        # with the run_id added on, sorted by root directory
        #
        # To get all that we must parse runinfo files and match them with demux files,
        # because demux files don't contain run-ids we need to match demux and runinfo
        # logs from the same directory, but find_log_files() does not guarantee order;
        # however it provides root dir, so we use that.
        demuxes_by_root = {f["root"]: f for f in self.find_log_files("bclconvert/demux")}
        runinfos_by_root = {f["root"]: f for f in self.find_log_files("bclconvert/runinfo")}
        qmetrics_by_root = {f["root"]: f for f in self.find_log_files("bclconvert/quality_metrics")}

        for root in runinfos_by_root.copy().keys():
            if root not in demuxes_by_root:
                log.error(f"Found RunInfo.xml file in {root} but no Demux Stats file, skipping")
                del runinfos_by_root[root]
                continue
        for root in demuxes_by_root.copy().keys():
            if root not in runinfos_by_root:
                log.error(f"Found Demux Stats file in {root} but no RunInfo.xml file, skipping")
                del demuxes_by_root[root]
                continue

        multiple_sequencing_runs = None
        last_run_id = None
        for root, demux in demuxes_by_root.items():
            runinfo = self._parse_single_runinfo_file(runinfos_by_root[root])
            demux["run_id"] = runinfo["run_id"]
            demux["cluster_length"] = runinfo["cluster_length"]
            qmetrics = qmetrics_by_root.get(root)
            if qmetrics:
                qmetrics["run_id"] = runinfo["run_id"]

            if last_run_id and runinfo["run_id"] != last_run_id:
                # this will mean we supress unknown reads, since we can't do a recalculation
                multiple_sequencing_runs = True
            last_run_id = runinfo["run_id"]

        return demuxes_by_root, qmetrics_by_root, multiple_sequencing_runs, last_run_id

    def _recalculate_undetermined(self, bclconvert_data, last_run_id):
        # We have to calculate "corrected" unknown read counts when parsing more than
        # one bclconvert run. To do this: add up all the reads in a lane that were
        # assigned to samples, then take the total reads in a lane (which is taken
        # _from the sum of all reads in a single file_), subtract the former from the
        # latter, and use that as "undetermined samples in lane."
        total_reads_per_lane = dict()
        for filename, lanedata in self.total_reads_in_lane_per_file.items():
            for lane_id, reads in lanedata.items():
                if lane_id not in total_reads_per_lane:
                    total_reads_per_lane[lane_id] = int()
                if total_reads_per_lane[lane_id] and reads != total_reads_per_lane[lane_id]:
                    log.error(
                        "Warning: different amounts of reads per lane across input files! "
                        "Cannot expect calculations to be accurate!"
                    )
                total_reads_per_lane[lane_id] = reads

        run_data = bclconvert_data[last_run_id]  # in this situation we have only one run id
        for lane_id, lane in run_data.items():
            determined_reads = 0
            for sample_id, sample in lane["samples"].items():
                determined_reads += sample["clusters"]
            self.per_lane_undetermined_reads[lane_id] = total_reads_per_lane[lane_id] - determined_reads

    def parse_demux_data(self, demux_file, bclconvert_data, num_demux_files):
        """
        Parse a bclconvert output stats csv, populate variables appropriately
        """
        filename = str(os.path.join(demux_file["root"], demux_file["fn"]))
        total_reads_in_lane = dict()

        run_data = bclconvert_data.get(demux_file["run_id"], dict())
        bclconvert_data[demux_file["run_id"]] = run_data
        with open(filename) as fh:
            reader: csv.DictReader = csv.DictReader(fh, delimiter=",")
            for row in reader:
                lane_id = f"L{row['Lane']}"
                lane = run_data.get(lane_id)
                if lane is None:
                    lane = dict(samples={}, cluster_length=demux_file["cluster_length"], **self._reads_dictionary())
                    run_data[lane_id] = lane
                    self.per_lane_undetermined_reads[lane_id] = 0

                sname = row["SampleID"]
                if sname != "Undetermined":
                    # Don't include undetermined reads at all in any of the calculations...
                    if sname not in lane["samples"]:
                        lane["samples"][sname] = self._reads_dictionary()
                        lane["samples"][sname]["filename"] = os.path.join(demux_file["root"], demux_file["fn"])
                        lane["samples"][sname]["samples"] = {}

                    sample = lane["samples"][sname]  # this sample in this lane

                    # total lane stats
                    lane["clusters"] += int(row["# Reads"])
                    lane["_calculated_yield"] += int(row["# Reads"]) * demux_file["cluster_length"]
                    lane["perfect_index_reads"] += int(row["# Perfect Index Reads"])
                    lane["one_mismatch_index_reads"] += int(row["# One Mismatch Index Reads"])
                    lane["basesQ30"] += int(row.get("# of >= Q30 Bases (PF)", "0"))  # Column only present pre v3.9.3

                    # stats for this sample in this lane
                    sample["clusters"] += int(row["# Reads"])
                    sample["_calculated_yield"] += int(row["# Reads"]) * demux_file["cluster_length"]
                    sample["perfect_index_reads"] += int(row["# Perfect Index Reads"])
                    sample["one_mismatch_index_reads"] += int(row["# One Mismatch Index Reads"])
                    sample["index"] = str(row["Index"])
                    try:
                        # Not all demux files have Sample_Project column
                        sample["sample_project"] = str(row["Sample_Project"])
                    except KeyError:
                        pass

                    # columns only present pre v3.9.3, after they moved to quality_metrics
                    sample["basesQ30"] += int(row.get("# of >= Q30 Bases (PF)", 0))
                    # Collecting to re-calculate mean_quality:
                    calculated_quality_score_sum = (
                        float(row.get("Mean Quality Score (PF)", 0)) * sample["_calculated_yield"]
                    )
                    sample["_calculated_quality_score_sum"] += calculated_quality_score_sum
                    lane["_calculated_quality_score_sum"] += calculated_quality_score_sum

                if lane_id not in total_reads_in_lane:
                    total_reads_in_lane[lane_id] = 0

                # Add up number of reads, regardless of undetermined or not
                total_reads_in_lane[lane_id] += int(row["# Reads"])

                if num_demux_files == 1 and sname == "Undetermined":
                    self.per_lane_undetermined_reads[lane_id] += int(row["# Reads"])

        self.total_reads_in_lane_per_file[filename] = total_reads_in_lane

    def parse_qmetrics_data(self, bclconvert_data, qmetrics_file):
        """
        Parse a bclconvert output stats CSV, populate variables appropriately
        """
        filename = str(os.path.join(qmetrics_file["root"], qmetrics_file["fn"]))
        self.total_reads_in_lane_per_file[filename] = dict()

        reader: csv.DictReader = csv.DictReader(open(filename), delimiter=",")
        for row in reader:
            run_data = bclconvert_data[qmetrics_file["run_id"]]
            lane_id = f"L{row['Lane']}"
            if lane_id not in run_data:
                log.warning(f"Found unrecognised lane {lane_id} in Quality Metrics file, skipping")
                continue
            lane = run_data[lane_id]
            sample = row["SampleID"]
            if sample != "Undetermined":  # don't include undetermined reads at all in any of the calculations...
                if sample not in run_data[lane_id]["samples"]:
                    log.warning(f"Found unrecognised sample {sample} in Quality Metrics file, skipping")
                    continue
                lane_sample = run_data[lane_id]["samples"][sample]  # this sample in this lane

                # Parse the stats that moved to this file in v3.9.3
                lane["yield"] += int(row["Yield"])
                lane["basesQ30"] += int(row["YieldQ30"])
                lane_sample["yield"] += int(row["Yield"])
                lane_sample["basesQ30"] += int(row["YieldQ30"])
                # Collecting to re-calculate mean_quality:
                lane["_quality_score_sum"] += float(row["QualityScoreSum"])
                lane_sample["_quality_score_sum"] += float(row["QualityScoreSum"])

    def _parse_top_unknown_barcodes(self, bclconvert_data, last_run_id):
        run_data = bclconvert_data[last_run_id]

        for unknown_barcode_file in self.find_log_files("bclconvert/unknown_barcodes", filehandles=True):
            barcode_reader = csv.DictReader(unknown_barcode_file["f"], delimiter=",")
            for unknown_barcode_row in barcode_reader:
                thislane = "L" + str(unknown_barcode_row["Lane"])
                if "top_unknown_barcodes" not in run_data[thislane]:
                    run_data[thislane]["top_unknown_barcodes"] = {}
                thisbarcode = str(unknown_barcode_row["index"]) + "-" + str(unknown_barcode_row["index2"])
                run_data[thislane]["top_unknown_barcodes"][thisbarcode] = int(unknown_barcode_row["# Reads"])

    @staticmethod
    def _total_reads_for_run(bclconvert_data, run_id):
        totalreads = 0
        for lane_id, lane in bclconvert_data[run_id].items():
            totalreads += lane["clusters"]
        return totalreads

    @staticmethod
    def _total_reads_all_runs(bclconvert_data):
        totalreads = 0
        for key, run_data in bclconvert_data.items():
            for lane_id, lane in run_data.items():
                totalreads += lane["clusters"]
        return totalreads

    def _set_lane_percentage_stats(self, data, cluster_length):
        try:
            data["percent_Q30"] = (float(data["basesQ30"]) / float(data["clusters"] * cluster_length)) * 100.0
        except ZeroDivisionError:
            data["percent_Q30"] = "NA"
        try:
            data["percent_perfectIndex"] = (float(data["perfect_index_reads"]) / float(data["clusters"])) * 100.0
        except ZeroDivisionError:
            data["percent_perfectIndex"] = "NA"
        try:
            data["percent_oneMismatch"] = float(data["one_mismatch_index_reads"]) / float(data["clusters"]) * 100.0
        except ZeroDivisionError:
            data["percent_oneMismatch"] = "NA"
        if self._get_genome_size():
            data["depth"] = float(data["basesQ30"]) / self._get_genome_size()
        else:
            data["depth"] = "NA"

    def _split_data_by_lane_and_sample(self, bclconvert_data) -> Tuple[Dict, Dict, Dict, Dict]:
        """
        Populate a collection of "stats across all lanes" and "stats across all samples"
        """
        bclconvert_by_lane = dict()
        bclconvert_by_sample = dict()
        count_by_sample_by_lane = defaultdict(lambda: defaultdict(int))  # counts only - used for a stacked bargraph
        source_files = dict()
        for run_id, r in bclconvert_data.items():
            # set stats for each lane (across all samples) in bclconvert_bylane dictionary
            for lane_id, lane in r.items():
                self._set_lane_percentage_stats(lane, lane["cluster_length"])

                lane_key_name = self.prepend_runid(run_id, lane_id)
                bclconvert_by_lane[lane_key_name] = {
                    "depth": lane["depth"],
                    "clusters": lane["clusters"],
                    "yield": lane["yield"] or lane["_calculated_yield"],
                    "perfect_index_reads": lane["perfect_index_reads"],
                    "one_mismatch_index_reads": lane["one_mismatch_index_reads"],
                    "basesQ30": lane["basesQ30"],
                    "percent_Q30": lane["percent_Q30"],
                    "percent_perfectIndex": lane["percent_perfectIndex"],
                    "percent_oneMismatch": lane["percent_oneMismatch"],
                    "top_unknown_barcodes": lane["top_unknown_barcodes"] if "top_unknown_barcodes" in lane else {},
                    "_quality_score_sum": lane["_quality_score_sum"] or lane["_calculated_quality_score_sum"],
                }

                # now set stats for each sample (across all lanes) in bclconvert_bysample dictionary
                for sample_id, sample in lane["samples"].items():
                    if sample_id not in bclconvert_by_sample:
                        bclconvert_by_sample[sample_id] = self._reads_dictionary()

                    s = bclconvert_by_sample[sample_id]
                    s["clusters"] += int(sample["clusters"])
                    s["yield"] += sample["yield"] or sample["_calculated_yield"]
                    s["perfect_index_reads"] += sample["perfect_index_reads"]
                    s["one_mismatch_index_reads"] += sample["one_mismatch_index_reads"]
                    s["basesQ30"] += sample["basesQ30"]
                    s["cluster_length"] = lane["cluster_length"]
                    s["_quality_score_sum"] += sample["_quality_score_sum"] or sample["_calculated_quality_score_sum"]
                    s["index"] = sample["index"]
                    try:
                        # Not all demux files have Sample_Project column
                        s["sample_project"] = sample["sample_project"]
                    except KeyError:
                        pass

                    if not self._get_genome_size():
                        s["depth"] = "NA"
                    else:
                        s["depth"] += float(sample["basesQ30"]) / self._get_genome_size()

                    if sample_id not in ["top_unknown_barcodes"]:
                        if sample_id not in source_files:
                            source_files[sample_id] = []
                        source_files[sample_id].append(sample["filename"])

                    count_by_sample_by_lane[sample_id][lane_key_name] += sample["clusters"]

        return bclconvert_by_lane, bclconvert_by_sample, count_by_sample_by_lane, source_files

    def sample_stats_table(self, bclconvert_data, bclconvert_by_sample):
        sample_stats_data = dict()
        total_reads = self._total_reads_all_runs(bclconvert_data)
        depth_available = False

        for sample_id, sample in bclconvert_by_sample.items():
            # Percent stats for bclconvert-bysample i.e. stats for sample across all lanes
            try:
                perfect_percent = f"{float(100.0 * sample['perfect_index_reads'] / sample['clusters']):.1f}"
            except ZeroDivisionError:
                perfect_percent = "0.0"
            try:
                one_mismatch_percent = "{0:.1f}".format(
                    float(100.0 * sample["one_mismatch_index_reads"] / sample["clusters"])
                )
            except ZeroDivisionError:
                one_mismatch_percent = "0.0"

            try:
                yield_q30_percent = f"{float(100.0 * (sample['basesQ30'] / sample['yield'])):.1f}"
            except ZeroDivisionError:
                yield_q30_percent = "0.0"  #

            try:
                percent_yield = (float(sample["yield"]) / float(total_reads * sample["cluster_length"])) * 100.0
            except ZeroDivisionError:
                percent_yield = "NA"

            try:
                percent_reads = (float(sample["clusters"]) / float(total_reads)) * 100.0
            except ZeroDivisionError:
                percent_reads = "NA"

            sample_stats_data[sample_id] = {
                "depth": sample["depth"],
                "basesQ30": sample["basesQ30"],
                "clusters": sample["clusters"],
                "percent_reads": percent_reads,
                "yield": sample["yield"],
                "percent_yield": percent_yield,
                "yield_q30_percent": yield_q30_percent,
                # "perfect_index": samle['perfect_index_reads'], # don't need these
                # "one_mismatch_index_reads": sample['one_mismatch_index_reads'],
                "perfect_pecent": perfect_percent,
                "one_mismatch_pecent": one_mismatch_percent,
                "mean_quality": sample.get("mean_quality"),
                "index": sample["index"],
            }
            if sample["depth"] != "NA":
                depth_available = True
            # Not all demux files have Sample_Project column
            if "sample_project" in sample:
                sample_stats_data[sample_id]["sample_project"] = sample["sample_project"]

        headers = {}
        if depth_available:
            headers["depth"] = {
                "title": "Coverage",
                "description": (
                    "Estimated sequencing depth based on the number of bases with quality score greater or equal to Q30, "
                    + (
                        f", assuming the genome size is {self._get_genome_size()} as provided in config"
                        if self._get_genome_size() is not None
                        else ""
                    )
                ),
                "min": 0,
                "suffix": "X",
                "scale": "BuPu",
            }

        headers["clusters"] = {
            "title": f"{config.read_count_prefix} Clusters",
            "description": "Total number of clusters (read pairs) for this sample as determined by bclconvert demultiplexing ({})".format(
                config.read_count_desc
            ),
            "scale": "Blues",
            "shared_key": "read_count",
        }
        headers["yield"] = {
            "title": f"Yield ({config.base_count_prefix})",
            "description": "Total number of bases for this sample as determined by bclconvert demultiplexing ({})".format(
                config.base_count_desc
            ),
            "scale": "Greens",
            "shared_key": "base_count",
        }
        headers["percent_reads"] = {
            "title": "% Clusters",
            "description": "Percentage of clusters (read pairs) for this sample in this run, as determined by bclconvert demultiplexing",
            "scale": "Blues",
            "max": 100,
            "min": 0,
            "suffix": "%",
        }
        headers["percent_yield"] = {
            "title": "% Yield",
            "description": "Percentage of sequenced bases for this sample in this run",
            "scale": "Greens",
            "max": 100,
            "min": 0,
            "suffix": "%",
        }
        headers["basesQ30"] = {
            "title": f"Bases ({config.base_count_prefix}) ≥ Q30 (PF)",
            "description": "Number of bases with a Phred score of 30 or higher, passing filter ({})".format(
                config.base_count_desc
            ),
            "scale": "Blues",
            "shared_key": "base_count",
        }
        headers["yield_q30_percent"] = {
            "title": "% Bases ≥ Q30 (PF)",
            "description": "Percent of bases with a Phred score of 30 or higher, passing filter ({})".format(
                config.base_count_desc
            ),
            "scale": "Greens",
            "max": 100,
            "min": 0,
            "suffix": "%",
        }
        headers["perfect_pecent"] = {
            "title": "% Perfect Index",
            "description": "Percent of reads with perfect index (0 mismatches)",
            "max": 100,
            "min": 0,
            "scale": "RdYlGn",
            "suffix": "%",
        }
        headers["one_mismatch_pecent"] = {
            "title": "% One Mismatch Index",
            "description": "Percent of reads with one mismatch index",
            "max": 100,
            "min": 0,
            "scale": "RdYlGn",
            "suffix": "%",
        }
        headers["mean_quality"] = {
            "title": "Mean Quality Score",
            "description": "Mean quality score of bases",
            "min": 0,
            "max": 40,
            "scale": "RdYlGn",
        }
        headers["index"] = {
            "title": "Index",
            "description": "Sample index",
            "scale": False,
            "hidden": True,
        }
        headers["sample_project"] = {
            "title": "Project",
            "description": "Sample project",
            "scale": False,
            "hidden": True,
        }

        # Table config
        table_config = {
            "namespace": "bclconvert",
            "id": "bclconvert-sample-stats-table",
            "table_title": "bclconvert Sample Statistics",
        }

        return table.plot(sample_stats_data, headers, table_config)

    def lane_stats_table(self, bclconvert_by_lane):
        depth_available = False
        for lane_id, lane in bclconvert_by_lane.items():
            try:
                yield_q30_percent = f"{float(100.0 * (lane['basesQ30'] / lane['yield'])):.1f}"
            except ZeroDivisionError:
                yield_q30_percent = "0.0"
            bclconvert_by_lane[lane_id]["yield_q30_percent"] = yield_q30_percent
            if lane["depth"] != "NA":
                depth_available = True

        headers = {}
        if depth_available:
            headers["depth-lane"] = {
                "title": "Coverage",
                "description": (
                    "Estimated sequencing depth based on the number of bases with quality score greater or equal to Q30"
                    + (
                        f", assuming the genome size is {self._get_genome_size()} as provided in config"
                        if self._get_genome_size() is not None
                        else ""
                    )
                ),
                "suffix": "X",
                "scale": "BuPu",
            }

        headers["reads-lane"] = {
            "title": f"{config.read_count_prefix} Clusters",
            "description": "Total number of clusters (read pairs) for this sample as determined by bclconvert demultiplexing ({})".format(
                config.read_count_desc
            ),
            "scale": "Blues",
            "shared_key": "read_count",
        }
        headers["yield-lane"] = {
            "title": f"Yield ({config.base_count_prefix})",
            "description": "Total number of bases for this sample as determined by bclconvert demultiplexing ({})".format(
                config.base_count_desc
            ),
            "scale": "Greens",
            "shared_key": "base_count",
        }
        headers["basesQ30-lane"] = {
            "title": f"Bases ({config.base_count_prefix}) ≥ Q30 (PF)",
            "description": "Number of bases with a Phred score of 30 or higher, passing filter ({})".format(
                config.base_count_desc
            ),
            "scale": "Blues",
            "shared_key": "base_count",
        }
        headers["yield_q30_percent-lane"] = {
            "title": "% Bases ≥ Q30 (PF)",
            "description": "Percent of bases with a Phred score of 30 or higher, passing filter",
            "max": 100,
            "min": 0,
            "scale": "Greens",
        }
        headers["perfect_index_reads-lane"] = {
            "title": f"{config.read_count_prefix} Perfect Index",
            "description": f"Reads with perfect index - 0 mismatches ({config.read_count_desc})",
            "scale": "Blues",
            "shared_key": "read_count",
        }

        headers["one_mismatch_index_reads-lane"] = {
            "title": f"{config.read_count_prefix} One Mismatch",
            "description": f"Reads with one mismatch index ({config.read_count_desc})",
            "scale": "Spectral",
            "shared_key": "read_count",
        }
        headers["percent_perfectIndex-lane"] = {
            "title": "% Perfect Index",
            "description": "Percent of reads with perfect index - 0 mismatches",
            "max": 100,
            "min": 0,
            "scale": "RdYlGn",
            "suffix": "%",
        }
        headers["percent_oneMismatch-lane"] = {
            "title": "% One Mismatch",
            "description": "Percent of reads with one mismatch",
            "max": 100,
            "min": 0,
            "scale": "RdYlGn",
            "suffix": "%",
        }
        headers["mean_quality-lane"] = {
            "title": "Mean Quality Score",
            "description": "Mean quality score of bases",
            "min": 0,
            "max": 40,
            "scale": "PiYG",
        }

        # Table config
        table_config = {
            "namespace": "bclconvert-lane",
            "id": "bclconvert-lane-stats-table",
            "table_title": "bclconvert Lane Statistics",
            "col1_header": "Run ID - Lane",
        }

        # new dict with matching keys for plotting (this avoids duplicate html id linting errors)
        bclconvert_bylane_foroutput = dict()
        for laneid, lanestats in bclconvert_by_lane.items():
            if laneid not in bclconvert_bylane_foroutput:
                bclconvert_bylane_foroutput[laneid] = dict()
            for key, value in lanestats.items():
                bclconvert_bylane_foroutput[laneid][key + "-lane"] = value

        return table.plot(bclconvert_bylane_foroutput, headers, table_config)

    @staticmethod
    def prepend_runid(runid, rest):
        return str(runid) + " - " + str(rest)

    def get_bar_data_from_counts(self, bclconvert_data, counts, last_run_id) -> Dict[str, Dict[str, int]]:
        # For per-lane stats we fetch undetermined reads, too.
        bar_data = {}
        for key, value in counts.items():
            bar_data[key] = {
                "perfect": value["perfect_index_reads"],
                "imperfect": value["clusters"] - value["perfect_index_reads"],
            }
            try:
                if key.startswith(
                    self.prepend_runid(last_run_id, "")  # this wont run in multiple sequencing run situations
                ):  # per-lane stats start with a prepended run id, this is a per-lane entry
                    this_lane_id = key.replace(self.prepend_runid(last_run_id, ""), "")
                    rundata = bclconvert_data[last_run_id]
                    if this_lane_id in rundata:  # this is definitely a lane
                        bar_data[key]["undetermined"] = self.per_lane_undetermined_reads[this_lane_id]
            except TypeError:
                # do nothing, there is no Undetermined - this will happen in case of multiple run ids
                pass

        return bar_data

    @staticmethod
    def get_bar_data_from_undetermined(flowcells):
        """
        Get data to plot for undetermined barcodes.
        """

        bar_data = defaultdict(dict)
        # get undetermined barcodes for each lanes
        for lane_id, lane in flowcells.items():
            try:
                for barcode, count in islice(lane["top_unknown_barcodes"].items(), 20):
                    bar_data[barcode][lane_id] = count
            except AttributeError:
                pass
            except KeyError:
                pass

        return {key: value for key, value in islice(bar_data.items(), 20)}



""" MultiQC functions to plot a heatmap """


import logging

from multiqc.utils import config
from multiqc.plots.plotly import heatmap

logger = logging.getLogger(__name__)

letters = "abcdefghijklmnopqrstuvwxyz"

# Load the template so that we can access its configuration
# Do this lazily to mitigate import-spaghetti when running unit tests
_template_mod = None


def get_template_mod():
    global _template_mod
    if not _template_mod:
        _template_mod = config.avail_templates[config.template].load()
    return _template_mod


def plot(data, xcats=None, ycats=None, pconfig=None):
    """Plot a 2D heatmap.
    :param data: List of lists, each a representing a row of values; or a dict of dicts
    :param xcats: Labels for x-axis
    :param ycats: Labels for y-axis. Defaults to same as x.
    :param pconfig: optional dict with config key:value pairs.
    :return: HTML and JS, ready to be inserted into the page
    """
    if pconfig is None:
        pconfig = {}

    # Allow user to overwrite any given config for this plot
    if "id" in pconfig and pconfig["id"] and pconfig["id"] in config.custom_plot_config:
        for k, v in config.custom_plot_config[pconfig["id"]].items():
            pconfig[k] = v

    if ycats is None:
        ycats = xcats

    # Make a plot
    mod = get_template_mod()
    if "heatmap" in mod.__dict__ and callable(mod.heatmap):
        try:
            return mod.heatmap(data, xcats, ycats, pconfig)
        except:  # noqa: E722
            if config.strict:
                # Crash quickly in the strict mode. This can be helpful for interactive
                # debugging of modules
                raise

    return heatmap.plot(data, pconfig, xcats, ycats)
